import numpy as np
from microquake.helpers.logging import logger

"""
    Place to keep trace transforms - mainly rotations to ENZ, PSVSH, etc.
"""


def rotate_to_P_SV_SH(st, cat, debug=False):

    fname = 'rotate_to_P_SV_SH'

    st_new = st.copy()

    event = cat[0]
    for arr in event.preferred_origin().arrivals:
        if arr.phase == 'S':
            continue

        pk = arr.pick_id.get_referred_object()
        sta = pk.waveform_id.station_code
        baz = arr.backazimuth
        az = arr.azimuth
        takeoff = arr.takeoff_angle
        inc_angle = arr.inc_angle

        if inc_angle is None:
            baz, inc_angle = event.preferred_origin(
            ).get_incidence_baz_angles(sta, arr.phase)
            inc_angle *= 180 / np.pi

        if inc_angle is None:
            logger.warning("%s: sta:%s [%s] has inc_angle=None --> skip "
                           "rotation!" % (fname, sta, arr.phase))
            continue

        trs = st_new.select(station=sta)
        if len(trs) == 3:

            cos_i = np.cos(inc_angle * np.pi/180.)
            sin_i = np.sin(inc_angle * np.pi/180.)
            cos_baz = np.cos(baz * np.pi/180.)
            sin_baz = np.sin(baz * np.pi/180.)

            col1 = np.array([cos_i, sin_i, 0.])
            col2 = np.array([-sin_i*sin_baz, cos_i*sin_baz, -cos_baz])
            col3 = np.array([-sin_i*cos_baz, cos_i*cos_baz, sin_baz])

            A = np.column_stack((col1,col2,col3))

            if debug:
                print("sta:%s az:%.1f baz:%.1f takeoff:%.1f inc:%.1f" % (sta, az, baz, takeoff, inc_angle))

            E = trs[0].data
            N = trs[1].data
            Z = trs[2].data
            D = np.row_stack((Z,E,N))

            foo = A @ D

            #if sta in ['59', '87']:
                #trs.plot()


            trs[0].data = foo[0,:]
            trs[1].data = foo[1,:]
            trs[2].data = foo[2,:]
            trs[0].stats.channel='P'
            trs[1].stats.channel='SV'
            trs[2].stats.channel='SH'

            '''
            P = trs[0].copy().trim(starttime = pk.time -.02, endtime=pk.time +.02)
            SV = trs[1].copy().trim(starttime = pk.time -.02, endtime=pk.time +.02)
            SH = trs[2].copy().trim(starttime = pk.time -.02, endtime=pk.time +.02)

            S = np.sqrt(SV.data**2 + SH.data**2)
            print(type(S))
            print(S)

            PtoS = np.var(P.data)/np.var(S)
            print(type(PtoS))
            print(PtoS)

            print("P_max:%g SV_max:%g SH_max:%g P/S:%f" % (np.max(np.abs(P.data)), np.max(np.abs(SV.data)), \
                                                    np.max(np.abs(SH.data), PtoS)))

            #if sta in ['59', '87']:
                #trs.plot()
            #exit()
            '''


        else:
            print("sta:%s --> only has n=%d traces --> can't rotate" % (sta, len(trs)))

    return st_new


def rotate_to_ENZ(st, inventory):
    st_new = st.copy()

    for sta in st_new.unique_stations():

        trs = st_new.select(station=sta)

        if not inventory.select(sta):
            logger.warning(f'missing station "{sta}" in inventory')
            continue

        # catching edge case when a uniaxial sensors contains three traces
        # with two traces containing only NaN.

        if len(trs) == 3:
            if np.any([np.all(np.isnan(trs[0].data)),
                       np.all(np.isnan(trs[1].data)),
                       np.all(np.isnan(trs[2].data))]):
                continue

        if len(trs) == 3 and len(inventory.select(sta).channels) == 3:

            try:
                col1 = inventory.get_channel(sta=sta, cha='X').cosines
                col2 = inventory.get_channel(sta=sta, cha='Y').cosines
                col3 = inventory.get_channel(sta=sta, cha='Z').cosines
            except AttributeError as err:
                logger.error(err)
                continue

            A = np.column_stack((col1,col2,col3))
            At = A.transpose()

            x = trs[0].data
            y = trs[1].data
            z = trs[2].data
            D = np.row_stack((x,y,z))

            foo = At @ D

            trs[0].data = foo[0,:]
            trs[1].data = foo[1,:]
            trs[2].data = foo[2,:]
            trs[0].stats.channel='E'
            trs[1].stats.channel='N'
            trs[2].stats.channel='Z'


    return st_new




import numpy as np
import matplotlib.pyplot as plt
from matplotlib.text import Text
from matplotlib.transforms import offset_copy
from microquake.core import event

from microquake.helpers.logging import logger


def guttenberg_richter(magnitudes, dates, bin_size=0.05, b_range=[-2.0, -0.5],
	magnitude_type='Moment magnitude', xlim=[-2.0, 1.0], **kwargs):
	"""
	produce a Guttenberg Richter plot from a list of magnitudes

	:param magnitudes: List of magnitudes
	:type magnitudes: list of float
	:param times: list of event time associated with the magnitude
	:type time: list of microquake.core.UTCDateTime
	:param bin_size: the width of bins
	:type bin_size: float
	:param b_range: range over which the b value is calculated an fit ([min, max])
	:type b_range: list containing two floats
	:param magnitude_type: Type of magnitude
	:type magnitude_type: str
	:param xlim: limits of the x axis
	:type xlim: list containing two floats
	"""

	mag = np.array(magnitudes)

	wdt = 15
	num_years = (np.max(dates) - np.min(dates)) / (24 * 3600 * 365.25)

	bins = np.arange(xlim[0], xlim[1], bin_size)
	hist = np.histogram(mag, bins=bins)
	hist = hist[0][::-1]
	bins = bins[::-1]
	bins = bins[1::]
	cum_hist = hist.cumsum()
	cum_hist = np.array([float(nb) for nb in cum_hist])
	cum_hist /= float(num_years)

	new_cum_yearly_rate = []
	for i in cum_hist:
	    new_cum_yearly_rate.append(i)

	log_cum_sum = np.log10(new_cum_yearly_rate)

	cum = np.zeros(bins.shape)
	min_mag = b_range[0]
	max_mag = b_range[1]
	indices = np.nonzero((bins >= min_mag) & (bins <= max_mag))[0]
	b, a = np.polyfit(bins[indices], log_cum_sum[indices], 1)

	mg = np.arange(min_mag, max_mag, 0.1)
	fitmg = b * mg + a

	# Tracer()()
	plt.semilogy(bins, new_cum_yearly_rate, 'k.', **kwargs)
	plt.semilogy(mg, 10 ** fitmg, 'k--', label='best fit ($%0.1f\,M_w + %0.1f$)' % (b, a), **kwargs)
	plt.xlabel('%s' % magnitude_type)
	plt.ylabel('Number of events/year')
	plt.xlim(xlim)
	plt.legend()
	# ylim = plt.ylim()
	# plt.ylim([0, 10**0])

	# plt.show()


class Plot(object):

	def __init__(self, data=None, picks=None, site=None):

		self.data = data
		self.picks = picks
		self.extraData = None
		self.style = "all"
		self.onlyIfHasPicks = True
		self.maxNumPlots = 999

		self.dataDirty = True
		self._plotData = None
		self.numPlots = 0
		self.numTraces = 0

		self.extraCaption = ""

	def close(self):

		# need to check that a plot has been created
		plt.close()

	def setData(self, data):

		for curTr, tr in enumerate(data):
			if curTr == 0:
				xmin = tr.stats.starttime
				xmax = tr.stats.endtime
			else:
				if tr.stats.starttime < xmin:
					xmin = tr.stats.starttime
				if tr.stats.starttime > xmax:
					xmax = tr.stats.endtime

		data.trim(starttime=xmin, endtime=xmax, pad=True, fill_value=0)

		self.data = data
		self.dataDirty = True

	def setPicks(self, picks):

		self.picks = picks
		self.dataDirty = True

	def _prepare_data(self):

		if not isinstance(self.picks, list):
			self.picks = [self.picks]

		self.data.detrend('demean')
		stations = np.unique([tr.stats.station for tr in self.data])

		self._plotData = []
		self.numTraces = 0
		for station in stations:
			curPicks = []
			# loop on picking stages
			for cat2 in self.picks:
				curStagePicks = []

				if isinstance(cat2, event.Event):
					evts = cat2
				elif isinstance(cat2, event.Catalog):
					if not cat2.events:
						continue

					evts = cat2.events
				else:
					continue

				if not isinstance(evts, list):
					evts = [evts]

				for evt in evts:
					if not evt['picks']:
						continue

					prevPicks = evt['picks']

					# find the existing P and S picks for the current station
					for pick in prevPicks:
						if pick['waveform_id'].station_code != station:
							continue

						curStagePicks.append(pick)

					curPicks.append(curStagePicks)

			if self.onlyIfHasPicks:
				numPicks = np.sum([len(a) for a in curPicks])
				if numPicks == 0:
					continue

			trs = self.data.select(station=station)
			# if len(trs) == 3:
				# trs = trs.rotate_P_S()

			self._plotData.append({'traces': trs, 'picks': curPicks})
			self.numTraces += len(trs)

		self.numPlots = 0
		if self.style == "all":
			self.numPlots = self.numTraces
			if self.numPlots > self.maxNumPlots:
				self.numPlots = self.maxNumPlots

		self.dataDirty = False

	def _make_plot(self, ax, plt_data, max_val, curPicks, starttime, cur_starttime, sr, transOffset, caption=None):

		npts = len(plt_data)

		t = np.array([(starttime + tmp).datetime for tmp in np.arange(0, npts) / sr])
		ax.plot(t, plt_data / max_val)
		plt.ylim([-1, 1])
		if caption:
			ax.text(.95, .9, caption, transform=ax.transAxes, va='top', ha='right', fontsize=10, backgroundcolor='white')

		for curStage, pickStage in enumerate(curPicks):
			for pick in pickStage:
				pick_sample = pick.time.datetime
				col = 'red' if pick['phase_hint'] == 'P' else 'black'
				# col = 'red'
				ax.axvline(pick_sample, c=col)

				snr = None
				for c in pick.comments:
					if 'SNR=' not in c.text:
						continue

					snr = c.text.split('=')[1]

				displayText = '%s%s' % (pick.phase_hint, curStage)
				if snr:
					displayText = '%s - %s' % (displayText, snr)

				label = Text(pick_sample,
							ax.get_ylim()[1] * .7, displayText,
							color=col, backgroundcolor='white', size=10, alpha=.8, transform=transOffset)
				ax.add_artist(label)

				if hasattr(pick, 'tt_residual'):
					pick_sample = (pick.time - pick.tt_residual).datetime
					ax.axvline(pick_sample, c='green')

	# This function supports plotting several stages of picking on the same graph.
	# Simply pass in an array of catalogs to see the pick progression from one stage to the other.
	def plot(self):

		if self.dataDirty:
			self._prepare_data()
		if (self.style == "all" and self.numPlots == 0) \
			or not self._plotData:
			logger.warning('No data to plot!')
			return

		fig = None

		if self.style == "all":
			if self.extraData:
				self.numPlots += 1

			fig = plt.figure(figsize=(10, 2 * self.numPlots), dpi=100)

		plotOffset = 0

		for curSt, t in enumerate(self._plotData):
			trs = t['traces']
			curPicks = t['picks']

			if self.style == "per_station":
				self.numPlots = len(trs)
				if self.extraData:
					self.numPlots += 1
				fig = plt.figure(figsize=(10, 5 * self.numPlots), dpi=100)

			starttime = trs[0].stats.starttime
			sr = trs[0].stats.sampling_rate
			cur_starttime = starttime
			transOffset = None

			curTr = 0

			for curTr, tr in enumerate(trs):
				curPlot = plotOffset + curTr
				if curPlot >= self.maxNumPlots:
					# finished
					return

				ax = plt.subplot(self.numPlots, 1, curPlot + 1)
				caption = '%s - %s' % (tr.stats.station, tr.stats.channel)
				if self.extraCaption:
					caption = '%s - %s' % (caption, self.extraCaption[curSt])

				transOffset = offset_copy(ax.transData, fig=fig,
	                            x=5, y=0, units='points')
				max_val = np.max(np.abs(tr.data))
				if max_val == 0:
					continue

				cur_starttime = tr.stats.starttime
				self._make_plot(ax, tr.data, max_val, curPicks, starttime, cur_starttime, sr, transOffset,
					caption=caption)
				plt.title('station %s' % tr.stats.station)

			if self.style == "per_station":
				if self.extraData:
					ax = plt.subplot(self.numPlots, 1, self.numPlots)
					max_val = np.max(np.abs(self.extraData[curSt]))

					self._make_plot(ax, self.extraData[curSt], max_val, curPicks, starttime, cur_starttime, sr, transOffset)


				self.show()
			else:
				plotOffset += curTr + 1

	def show(self):

		plt.show()

	def saveFig(self, outFile=""):

		plt.tight_layout()
		plt.savefig(outFile, dpi=100, bbox_inches='tight')




""" Waveform amplitude measurements

    This module contains a collection of functions for making
    measurements on the velocity and displacement waveforms
    that are later used to calculate moment magnitude, focal mechanism, etc

"""

# default logger
from microquake.helpers.logging import logger

import matplotlib.pyplot as plt
import numpy as np
from scipy.fftpack import rfft

from microquake.core.data.inventory import get_sensor_type_from_trace
from microquake.core.event import Pick
from microquake.core.stream import Stream
from microquake.core.util.tools import copy_picks_to_dict
from microquake.waveform.parseval_utils import npow2, unpack_rfft
from microquake.waveform.pick import calculate_snr


def measure_pick_amps(st_in, cat, phase_list=None,
                      triaxial_only=False,
                      **kwargs):
    """
    Attempt to measure velocity pulse parameters (polarity, peak vel, etc)
      and displacement pulse parameters (pulse width, area)
      for each arrival for each event preferred_origin in cat

    Measures are made on individual traces, saved to arrival.traces[trace id],
      and later combined to one measurement per arrival
      and added to the *arrival* extras dict

    :param st_in: velocity traces
    :type st_in: obspy.core.Stream or microquake.core.Stream
    :param cat: obspy.core.event.Catalog
    :type cat: list of obspy.core.event.Events or microquake.core.event.Events
    :param phase_list: ['P'], ['S'], or ['P', 'S'] - list of arrival phases
    to process
    :type phase_list: list
    :param triaxial_only: if True --> only keep 3-component observations (
    disp area) in arrival dict
    :type triaxial_only: boolean
    """

    fname = "measure_pick_amps"

    st = st_in.copy()

    cat = measure_velocity_pulse(st, cat, phase_list=phase_list, **kwargs)

    debug = False

    if 'debug' in kwargs:
        debug = kwargs['debug']

    cat = measure_displacement_pulse(st, cat, phase_list=phase_list,
                                             debug=debug)

    # Combine individual trace measurements (peak_vel, dis_pulse_area, etc)
    #    into one measurement per arrival:

    for event in cat:
        for phase in phase_list:
            origin = event.preferred_origin() if event.preferred_origin() \
                else event.origins[0]

            for arr in origin.arrivals:
                pk = Pick(arr.get_pick())

                if not pk:
                    continue
                sta = pk.get_sta()

                if arr.traces is not None:
                    dis_area = []
                    dis_width = []

                    for tr_id, v in arr.traces.items():

                        if v['polarity'] != 0 and tr_id[-1].upper() in ['Z',
                                                                        'P']:
                            arr.polarity = v['polarity']
                            arr.t1 = v['t1']
                            arr.t2 = v['t2']
                            arr.peak_vel = v['peak_vel']
                            arr.tpeak_vel = v['tpeak_vel']
                            arr.pulse_snr = v['pulse_snr']

                        if v['peak_dis'] is not None and tr_id[-1].upper() \
                                in ['Z', 'P']:
                            arr.peak_dis = v['peak_dis']
                            arr.max_dis = v['max_dis']
                            arr.tpeak_dis = v['tpeak_dis']
                            arr.tmax_dis = v['tmax_dis']

            # But average vector quantities distributed over components:
                        if v['dis_pulse_area'] is not None:
                            dis_area.append(v['dis_pulse_area'])

                        if v['dis_pulse_width'] is not None:
                            dis_width.append(v['dis_pulse_width'])

            # Here is where you could impose triaxial_only requirement
            #  but this will filter out not only 1-chan stations, but
            #  any stations where peak finder did not locate peak on
            #  *all* 3 channels
                    # if triaxial_only and len(dis_area) == 3:

                    if triaxial_only and len(dis_area) == 3:
                        arr.dis_pulse_area = np.sqrt(np.sum(np.array(
                            dis_area) ** 2))
                    elif len(dis_area) > 0:
                        arr.dis_pulse_area = np.sqrt(np.sum(np.array(
                            dis_area) ** 2))

                    if len(dis_width) > 0:
                        arr.dis_pulse_width = np.mean(dis_width)

                else:
                    pass

    return cat.copy()


def measure_velocity_pulse(st,
                           cat,
                           phase_list=None,
                           pulse_min_width=.005,
                           pulse_min_snr_P=7,
                           pulse_min_snr_S=5,
                           debug=False,
                           ):
    """
    locate velocity pulse (zero crossings) near pick and measure peak amp,
        polarity, etc on it

    All measurements are added to the *arrival* extras dict

    :param st: velocity traces
    :type st: obspy.core.Stream or microquake.core.Stream
    :param cat: obspy.core.event.Catalog
    :type cat: list of obspy.core.event.Events or microquake.core.event.Events
    :param phase_list: ['P'], ['S'], or ['P', 'S'] - list of arrival phases
    to process
    :type phase_list: list
    :param pulse_min_width: Measured first pulse must be this wide to be
    retained
    :type pulse_min_width: float
    :param pulse_min_snr_P: Measure first P pulse must have snr greater
    than this
    :type pulse_min_snr_P: float
    :param pulse_min_snr_S: Measure first S pulse must have snr greater
    than this
    :type pulse_min_snr_S: float
    """

    fname = 'measure_velocity_pulse'

    traces_info = []

    if phase_list is None:
        phase_list = ['P']

    # Average of P,S min snr used for finding zeros
    min_pulse_snr = int((pulse_min_snr_P + pulse_min_snr_S)/2)

    for event in cat:
        origin = event.preferred_origin() if event.preferred_origin() else \
            event.origins[-1]
        arrivals = origin.arrivals

        for arr in arrivals:

            phase = arr.phase

            if phase not in phase_list:
                continue

            pk = Pick(arr.get_pick())

            if pk is None:
                logger.error("%s: arr pha:%s id:%s --> Lost reference to "
                             "pick id:%s --> SKIP" %
                             (fname, arr.phase, arr.resource_id.id,
                              arr.pick_id.id))

                continue
            sta = pk.get_sta()

            trs = st.select(station=sta)

            if trs is None:
                logger.warning("%s: sta:%s has a [%s] arrival but no trace "
                               "in stream --> Skip" % (fname, sta, arr.phase))
                continue

            sensor_type = get_sensor_type_from_trace(trs[0])

            if sensor_type != "VEL":
                logger.info("%s: sta:%s sensor_type != VEL --> Skip"
                            % (fname, sta))
                continue

            arr.traces = {}

            for tr in trs:
                try:
                    tr.detrend("demean").detrend("linear")
                except Exception as e:
                    print(e)

                    continue
                data = tr.data.copy()
                ipick = int((pk.time - tr.stats.starttime) *
                            tr.stats.sampling_rate)

                polarity, vel_zeros = _find_signal_zeros(
                    tr, ipick,
                    nzeros_to_find=3,
                    min_pulse_width=pulse_min_width,
                    min_pulse_snr=min_pulse_snr,
                    debug=debug
                )

                dd = {}
                dd['polarity'] = 0
                dd['t1'] = None
                dd['t2'] = None
                dd['peak_vel'] = None
                dd['tpeak_vel'] = None
                dd['pulse_snr'] = None

                # A good pick will have the first velocity pulse located
                # between i1 and i2

                if vel_zeros is not None:
                    i1 = vel_zeros[0]
                    i2 = vel_zeros[1]
                    t1 = tr.stats.starttime + float(i1 * tr.stats.delta)
                    t2 = tr.stats.starttime + float(i2 * tr.stats.delta)

                    ipeak, peak_vel = _get_peak_amp(tr, i1, i2)
                    tpeak = tr.stats.starttime + float(ipeak * tr.stats.delta)

                    noise_npts = int(.01 * tr.stats.sampling_rate)
                    noise_end = ipick - int(.005 * tr.stats.sampling_rate)
                    noise = data[noise_end - noise_npts: noise_end]
                    noise1 = np.abs(np.mean(noise))
                    noise2 = np.abs(np.median(noise))
                    noise3 = np.abs(np.std(noise))
                    noise_level = np.max([noise1, noise2, noise3])

                    pulse_snr = np.abs(peak_vel/noise_level)
                    pulse_width = float((i2-i1)*tr.stats.delta)

                    pulse_thresh = pulse_min_snr_P

                    if phase == 'S':
                        pulse_thresh = pulse_min_snr_S

                    if pulse_snr < pulse_thresh:
                        logger.debug("%s: tr:%s pha:%s t1:%s t2:%s "
                                     "pulse_snr=%.1f < thresh" %
                                     (fname, tr.get_id(), phase, t1, t2,
                                      pulse_snr))
                        polarity = 0

                    if pulse_width < pulse_min_width:
                        logger.debug("%s: tr:%s pha:%s t1:%s t2:%s "
                                     "pulse_width=%f < %f" %
                                     (fname, tr.get_id(), phase, t1, t2,
                                      pulse_width, pulse_min_width))
                        polarity = 0

                    dd['polarity'] = polarity
                    dd['peak_vel'] = peak_vel
                    dd['tpeak_vel'] = tpeak
                    dd['t1'] = t1
                    dd['t2'] = t2
                    dd['pulse_snr'] = pulse_snr

                else:
                    logger.debug("%s: Unable to locate zeros for tr:%s pha:%s"
                                 % (fname, tr.get_id(), phase))

                arr.traces[tr.get_id()] = dd
                dd['trace_id'] = tr.get_id()
                dd['arrival_id'] = arr.resource_id
                dd['event_id'] = event.resource_id
                dd['origin_id'] = origin.resource_id
                traces_info.append(dd)

            # Process next phase in phase_list

        # Process tr in st

    # Process next event in cat

    return cat


def measure_displacement_pulse(st,
                               cat,
                               phase_list=None,
                               debug=False):
    """
    measure displacement pulse (area + width) for each pick on each arrival
        as needed for moment magnitude calculation

    All measurements are added to the *arrival* extras dict

    :param st: velocity traces
    :type st: obspy.core.Stream or microquake.core.Stream
    :param cat: obspy.core.event.Catalog
    :type list: list of obspy.core.event.Events or microquake.core.event.Events
    """

    fname = 'measure_displacement_pulse'

    if phase_list is None:
        phase_list = ['P']

    traces_info = []
    for event in cat:
        origin = event.preferred_origin() if event.preferred_origin() else \
            event.origins[0]
        arrivals = origin.arrivals

        for arr in arrivals:

            phase = arr.phase

            if phase not in phase_list:
                continue

            pk = Pick(arr.get_pick())

            if pk is None:
                logger.error("%s: arr pha:%s id:%s --> Lost reference to "
                             "pick id:%s --> SKIP" %
                             (fname, arr.phase, arr.resource_id.id,
                              arr.pick_id.id))

                continue
            sta = pk.get_sta()

            trs = st.select(station=sta)

            if trs is None:
                logger.warning("%s: sta:%s has a [%s] arrival but no trace "
                               "in stream --> Skip" %
                               (fname, sta, arr.phase))

                continue

            sensor_type = get_sensor_type_from_trace(trs[0])

            if sensor_type != "VEL":
                logger.info("%s: sta:%s sensor_type != VEL --> Skip"
                            % (fname, sta))

                continue

            for tr in trs:
                try:
                    tr_dis = tr.copy().detrend("demean").detrend("linear")
                    tr_dis.integrate().detrend("linear")
                except Exception as e:
                    print(e)

                    continue
                tr_dis.stats.channel = "%s.dis" % tr.stats.channel

                dd = {}
                dd['peak_dis'] = None
                dd['max_dis'] = None
                dd['tpeak_dis'] = None
                dd['tmax_dis'] = None
                dd['dis_pulse_width'] = None
                dd['dis_pulse_area'] = None

                tr_dict = arr.traces[tr.get_id()]

                polarity = tr_dict['polarity']
                t1 = tr_dict.get('t1', None)
                t2 = tr_dict.get('t2', None)

                if polarity != 0:

                    if t1 is None or t2 is None:
                        logger.error("%s: t1 or t2 is None --> You shouldn't "
                                     "be here!" % (fname))

                        continue

                    i1 = int((t1 - tr.stats.starttime) *
                             tr.stats.sampling_rate)
                    i2 = int((t2 - tr.stats.starttime) *
                             tr.stats.sampling_rate)

                    ipick = int((pk.time - tr.stats.starttime) *
                                tr.stats.sampling_rate)

                    icross = i2
                    tr_dis.data = tr_dis.data - tr_dis.data[i1]
                    # tr_dis.data = tr_dis.data - tr_dis.data[ipick]

                    dis_polarity = np.sign(tr_dis.data[icross])
                    pulse_width, pulse_area = _get_pulse_width_and_area(
                        tr_dis, i1, icross)

                    npulse = int(pulse_width * tr.stats.sampling_rate)

                    max_pulse_duration = .08
                    nmax_len = int(max_pulse_duration * tr.stats.sampling_rate)

                    if pulse_width != 0:

                        ipeak, peak_dis = _get_peak_amp(tr_dis, ipick,
                                                        ipick + npulse)
                        # max_dis = max within max_pulse_duration of pick time
                        imax, max_dis = _get_peak_amp(tr_dis, ipick,
                                                      ipick + nmax_len)

                        tmax_dis = tr.stats.starttime + float(imax *
                                                              tr.stats.delta)
                        tpeak_dis = tr.stats.starttime + float(ipeak *
                                                               tr.stats.delta)
                        tcross_dis = pk.time + pulse_width

                        dd['peak_dis'] = peak_dis
                        dd['max_dis'] = max_dis
                        dd['tpeak_dis'] = tpeak_dis
                        dd['tmax_dis'] = tmax_dis
                        dd['dis_pulse_width'] = pulse_width
                        dd['dis_pulse_area'] = pulse_area

                        if debug:
                            logger.debug("[%s] Dis pol=%d tpick=%s" %
                                         (phase, dis_polarity, pk.time))
                            logger.debug("              tpeak=%s "
                                         "peak_dis=%12.10g" %
                                         (tpeak_dis, peak_dis))
                            logger.debug("             tcross=%s" % tcross_dis)
                            logger.debug("               tmax=%s "
                                         "max_dis=%12.10g" %
                                         (tmax_dis, max_dis))
                            logger.debug("    dis pulse width=%.5f"
                                         % pulse_width)
                            logger.debug("    dis pulse  area=%12.10g"
                                         % pulse_area)

                    else:
                        logger.warning("%s: Got pulse_width=0 for tr:%s pha:%s"
                                       % (fname, tr.get_id(), phase))

                arr.traces[tr.get_id()] = dict(tr_dict, **dd)

                dd['trace_id'] = tr.get_id()
                dd['arrival_id'] = arr.resource_id
                dd['event_id'] = event.resource_id
                dd['origin_id'] = origin.resource_id
                traces_info.append(dd)


            # Process next tr in trs

        # Process next arr in arrivals

    # Process next event in cat

    return cat


def _find_signal_zeros(tr, istart, max_pulse_duration=.1, nzeros_to_find=3,
                       second_try=False, debug=False,
                       min_pulse_width=.00167, min_pulse_snr=5):
    """
    Locate zero crossing of velocity trace to locate first pulse(s)
    All measurements are added to the *arrival* extras dict

    :param tr: Individual velocity trace
    :type tr: obspy.core.Trace or microquake.core.Trace
    :param istart: index in trace.data to start searching from (eg =pick index)
    :type istart: int
    :param max_pulse_duration: maximum search window (in seconds) for end
                               of pulse
    :type max_pulse_duration: float
    :param nzeros_to_find: Number of zero crossings to find
    :type nzeros_to_find: int
    :param second_try: If true then iterate once over early/late pick to locate
                       first pulse
    :type second_try: boolean
    :param min_pulse_width: minimum allowable pulse width (in seconds)
    :type min_pulse_width: float
    :param min_pulse_snr: minimum allowable pulse snr
    :type min_pulse_snr: float
    :param debug: If true then output/plot debug msgs
    :type debug: boolean

    :returns: first_sign, zeros
    :rtype: int, np.array
    """

    fname = '_find_signal_zeros'

    data = tr.data
    sign = np.sign(data)

    i1 = -9

    if second_try:
        logger.debug("%s: This is the second try!" % fname)

    noise_tlen = .05
    noise_npts = int(noise_tlen * tr.stats.sampling_rate)
    noise_end = istart - int(.005 * tr.stats.sampling_rate)
    # noise_level = np.mean(data[noise_end - noise_npts: noise_end])
    noise = data[noise_end - noise_npts: noise_end]
    noise1 = np.abs(np.mean(noise))
    noise2 = np.abs(np.median(noise))
    noise3 = np.abs(np.std(noise))
    noise_level = np.max([noise1, noise2, noise3])
    noise_level = noise1

    # pick_snr = np.abs(data[istart]/noise_level)

    nmax_look = int(max_pulse_duration * tr.stats.sampling_rate)

    # Just used for debug
    pick_time = tr.stats.starttime + float(istart*tr.stats.delta)

# Stage 0: Take polarity sign (s0) from first data point after
#          after istart (=ipick) with SNR >= thresh * noise_level

    s0 = 0
    i0 = 0
    snr_thresh = 10.

    iend = istart + nmax_look
    if iend > len(data):
        iend = len(data)

    for i in range(istart, iend):
        if np.abs(data[i]) >= snr_thresh * np.abs(noise_level):
            s0 = sign[i]
            i0 = i

            break

# Stage 1: Back up from this first high SNR point to find the earliest point
#          with the *same* polarity.  Take this as i1 = pulse start
    i1 = i0
    s1 = s0
    snr_scale = 1.4

    if s0 and i0:
        for i in range(i0, istart - 3, -1):
            snr = np.abs(data[i]/noise_level)

            if sign[i] == s0 and snr >= snr_scale:
                # print("  sign matches --> set i1=i=%d" % i)
                i1 = i
            else:
                # print("  sign NO match --> break")

                break

    if i1 <= 0:
        logger.debug("%s: tr:%s pick_time:%s Didn't pass first test" %
                     (fname, tr.get_id(), pick_time))
        # tr.plot()

        return 0, None

    first_sign = s1


# Stage 2: Find the first zero crossing after this
#          And iterate for total of nzeros_to_find subsequent zeros

    zeros = np.array(np.zeros(nzeros_to_find,), dtype=int)
    zeros[0] = i1

    t1 = tr.stats.starttime + float(i1)*tr.stats.delta
    # print("i1=%d --> t=%s" % (i1, t1))


# TODO: Need to catch flag edge cases where we reach end of range with
#       no zero set!
    for j in range(1, nzeros_to_find):
        # for i in range(i1, i1 + 200):

        for i in range(i1, i1 + nmax_look):
            # print("j=%d i=%d sign=%d" % (j,i,sign[i]))

            if sign[i] != s1:
                # half_per = float( (i - i1) * tr.stats.delta)
                # f = .5 / half_per
                # ipeak,peak = get_peak_amp(tr, i1, i)
                # print("sign:[%2s] t1:%s - t2:%s (T/2:%.6f f:%f) peak:%g" % \
                # (s1, t1, t2, half_per, f, peak))
                i1 = i
                s1 = sign[i]
                zeros[j] = i

                break
    t1 = tr.stats.starttime + float(zeros[0])*tr.stats.delta
    t2 = tr.stats.starttime + float(zeros[1])*tr.stats.delta

# At this point, first (vel) pulse is located between zeros[0] and zeros[1]
    pulse_width = float(zeros[1]-zeros[0]) * tr.stats.delta
    ipeak, peak_vel = _get_peak_amp(tr, zeros[0], zeros[1])
# noise_level defined this way is just for snr comparison
    noise_level = np.max([noise1, noise2, noise3])
    pulse_snr = np.abs(peak_vel / noise_level)

    logger.debug("find_zeros: sta:%s cha:%s First pulse t1:%s t2:%s ["
                 "polarity:%d] pulse_width:%f peak:%g snr:%f" %
                 (tr.stats.station, tr.stats.channel, t1, t2, first_sign,
                  pulse_width, peak_vel, pulse_snr))

# Final gate = try to catch case of early pick on small bump preceding main
#              arrival move istart to end of precursor bump and retry
    if ((pulse_width < min_pulse_width or pulse_snr < min_pulse_snr)
            and not second_try):

        logger.debug("Let's RUN THIS ONE AGAIN ============== tr_id:%s" %
                     tr.get_id())
    # if pulse_width < min_pulse_width and not second_try:
        istart = zeros[1]

        return _find_signal_zeros(tr, istart,
                                  max_pulse_duration=max_pulse_duration,
                                  nzeros_to_find=nzeros_to_find,
                                  second_try=True)

    # if debug:
        # tr.plot()

    return first_sign, zeros


def _get_peak_amp(tr, istart, istop):
    """
    Measure peak (signed) amplitude between istart and istop on trace
    :param tr: velocity trace
    :type tr: obspy.core.trace.Trace or microquake.core.Trace
    :param istart: pick index in trace
    :type istart: int
    :param istop: max index in trace to search
    :type istart: int
    :returns: imax, amp_max: index + value of max
    :rtype: int, float
    """

    abs_max = -1e12

    if istop < istart:
        logger.error("_get_peak_amp: istart=%d < istop=%d !" % (istart, istop))
        exit()

    for i in range(istart, istop):
        if np.abs(tr.data[i]) >= abs_max:
            abs_max = np.abs(tr.data[i])
            imax = i

    return imax, tr.data[imax]


def _get_pulse_width_and_area(tr, ipick, icross, max_pulse_duration=.08):
    """
    Measure the width & area of the arrival pulse on the displacement trace
    Start from the displacement peak index (=icross - location of first zero
            crossing of velocity)

    :param tr: displacement trace
    :type tr: obspy.core.trace.Trace or microquake.core.Trace
    :param ipick: index of pick in trace
    :type ipick: int
    :param icross: index of first zero crossing in corresponding velocity trace
    :type icross: int
    :param max_pulse_duration: max allowed duration (sec) beyond pick to search
                               for zero crossing of disp pulse
    :type max_pulse_duration: float

    return pulse_width, pulse_area
    :returns: pulse_width, pulse_area: Returns the width and area of the
                                       displacement pulse
    :rtype: float, float
    """

    fname = '_get_pulse_width_and_area'

    data = tr.data
    sign = np.sign(data)

    nmax = int(max_pulse_duration * tr.stats.sampling_rate)
    iend = ipick + nmax

    epsilon = 1e-10

    if icross >= iend:
       i = iend - 1

    for i in range(icross, iend):
        diff = np.abs(data[i] - data[ipick])

        if diff < epsilon or sign[i] != sign[icross]:
            break

        if i == iend - 1:
            logger.info("%s: Unable to locate termination of displacement "
                        "pulse for tr:%s!" % (fname, tr.get_id()))

            return 0, 0

    istop = i
    pulse_width = float(istop - ipick) * tr.stats.delta
    pulse_area = np.trapz(data[ipick:istop], dx=tr.stats.delta)

    return pulse_width, pulse_area


def set_pick_snrs(st, picks, pre_wl=.03, post_wl=.03):
    """
    This function sets the pick snr on each individual trace
     (vs the snr_picker which calculates snr on the composite trace)

    The resulting snr is stored in the tr.stats[key] dict
    where key = {'P_arrival', 'S_arrival'}

    :param st: traces
    :type st: either obspy.core.Stream or microquake.core.Stream
    :param picks: P & S picks
    :type list: list of either obspy or microquake picks
    :param pre_wl: pre pick window for noise calc
    :type float:
    :param post_wl: post pick window for signal calc
    :type float:
    """

    pick_dict = copy_picks_to_dict(picks)

    for tr in st:
        sta = tr.stats.station

        if sta in pick_dict:
            for phase in pick_dict[sta]:
                pick_time = pick_dict[sta][phase].time

                if phase == 'S':
                    snr = calculate_snr(Stream(traces=[tr]), pick_time,
                                        pre_wl=pre_wl, post_wl=post_wl)
                else:
                    snr = calculate_snr(Stream(traces=[tr]), pick_time,
                                        pre_wl=pre_wl, post_wl=post_wl)
                key = "%s_arrival" % phase

                if key not in tr.stats:
                    tr.stats[key] = {}
                tr.stats[key]['snr'] = snr
        else:
            logger.warning("set_pick_snrs: sta:%s not in pick_dict" % sta)

    return


def calc_velocity_flux(st_in,
                       cat,
                       inventory,
                       phase_list=None,
                       use_fixed_window=True,
                       pre_P=.01,
                       P_len=.05,
                       pre_S=.01,
                       S_len=.1,
                       Q=1e12,
                       correct_attenuation=False,
                       triaxial_only=True,
                       debug=False):
    """
    For each arrival (on phase_list) calculate the velocity flux using
        the corresponding traces and save to the arrival.vel_flux to
        be used in the calculation of radiated seismic energy

    :param st_in: velocity traces
    :type st_in: obspy.core.Stream or microquake.core.Stream
    :param cat: obspy.core.event.Catalog
    :type cat: list of obspy.core.event.Events or microquake.core.event.Events
    :param phase_list: ['P'], ['S'], or ['P', 'S'] - list of arrival phases
    to process
    :type phase_list: list
    :param triaxial_only: if True --> only calc flux for 3-comp stations
    :type triaxial_only: boolean
    :param Q: Anelastic Q to use for attenuation correction to flux
    :type Q: float
    :param correct_attenuation: if True, scale spec by e^-pi*f*travel-time/Q
    before summing
    :type correct_attenuation: boolean
    """

    fname = "calc_velocity_flux"

    if phase_list is None:
        phase_list = ['P', 'S']

    # MTH: Try to avoid copying cat - it causes the events to lose their
    # link to preferred_origin!
    # cat = cat_in.copy()
    # Defensive copy - not currently needed since we only trim on copies,
    # still ...
    st = st_in.copy().detrend('demean').detrend('linear')

    for event in cat:
        origin = event.preferred_origin() if event.preferred_origin() else \
            event.origins[0]

        for arr in origin.arrivals:

            phase = arr.phase

            if phase not in phase_list:
                continue

            pick = Pick(arr.get_pick())

            if pick is None:
                logger.error("%s: arr pha:%s id:%s --> Lost reference to "
                             "pick id:%s --> SKIP" %
                             (fname, arr.phase, arr.resource_id.id,
                              arr.pick_id.id))

                continue
            sta = pick.get_sta()

            trs = st.select(station=sta)

            if trs is None:
                logger.warning("%s: sta:%s has a [%s] arrival but no trace "
                               "in stream --> Skip" %
                               (fname, sta, phase))

                continue

            if triaxial_only and len(trs) != 3:
                logger.info("%s: sta:%s is not 3-comp --> Skip" % (fname, sta))

                continue

            sensor_type = get_sensor_type_from_trace(trs[0])

            if sensor_type != "VEL":
                logger.info("%s: sta:%s sensor_type != VEL --> Skip" % (
                    fname, sta))

                continue

            if use_fixed_window:
                if phase == 'P':
                    pre = pre_P
                    win_secs = P_len
                else:
                    pre = pre_S
                    win_secs = S_len

                starttime = pick.time - pre
                endtime = starttime + win_secs

            not_enough_trace = False

            for tr in trs:
                if starttime < tr.stats.starttime or endtime > \
                        tr.stats.endtime:
                    logger.warning("%s: sta:%s pha:%s tr:%s is too short to "
                                   "trim --> Don't use" %
                                   (fname, sta, phase, tr.get_id()))
                    not_enough_trace = True

                    break

            if not_enough_trace:
                continue

            tr3 = trs.copy()

            tr3.trim(starttime=starttime, endtime=endtime)
            dt = tr3[0].stats.delta

            # flux_t = np.sum( [tr.data**2 for tr in tr3]) * dt

            tsum = 0

            for tr in tr3:
                tsum += np.sum(tr.data**2)*dt

            if not correct_attenuation:
                flux = tsum
                fsum = None

        # The only reason to do this in the freq domain is if we
        #    want to apply attenuation correction
            else:
                travel_time = pick.time - origin.time

                fsum = 0.

        # exp(pi * f * (R/v) * 1/Q) grows so fast with freq that it's out of
                # control above f ~ 1e3 Hz
        # We could design Q = Q(f) - e.g., make Q grow fast with f to
                # counteract this.
        # Alternatively, we limit the freq range of the (
                # attenuation-corrected) energy calc:
        #   To compare the t/f calcs using Parseval's:
        #   1. Set Q to something like 1e12 in the settings
        #   2. Set fmin=0 so that the low freqs are included in the summationi
                sensor_response = inventory.select(arr.get_sta())
                poles = np.abs(sensor_response[0].response.get_paz().poles)
                fmin = np.min(poles) / (2 * np.pi)
                fmax = 500. # looking at the response in the frequency
                # domain, beyond 500 Hz, the response is dominated by the
                # brown environmental noise.
        # In addition, it's necessary to locate fmin/fmax for each arrival
        # based on the min/max freqs where the velocity spec exceeds the
                # noise spec.
        # Thus, we can't actually include all the radiated energy, just the
                # energy above the noise level
        #   so these estimates will likely be low
                fmin = arr.fmin
                fmax = arr.fmax

                if fmax/fmin < 3:
                    logger.info("%s: sta:%s fmin:%.1f fmax:%.1f too "
                                "narrowband --> Skip"
                                % (fname, sta, fmin, fmax))

                    continue

                for tr in tr3:
                    data = tr.data
                    nfft = 2 * npow2(data.size)
                    df = 1./(dt * float(nfft))    # df is same as for
                    y, freqs = unpack_rfft(rfft(data, n=nfft), df)
                    y *= dt
                    y[1:-1] *= np.sqrt(2.)

                    tstar = travel_time / Q

                    index = [(freqs >= fmin) & (freqs <= fmax)]
                    freqs = freqs[index]
                    y = y[index]
                    fsum += np.sum(np.abs(y) * np.abs(y) * np.exp(
                        2. * np.pi * freqs * tstar)) * df

                print("arr sta:%s [%s] tsum=%g fsum=%g fmin=%f fmax=%f" %
                      (sta, arr.phase, tsum, fsum, fmin, fmax))
                # exit()

                flux = fsum

            # Note we are saving the "flux" but it has not (yet) been scaled
            # by rho*vel. Instead it is just the sum of the integrals of the
            # component velocities squared for the corresponding arrival and
            # scaling is done in calc_energy(..)

            # arr.vel_flux = flux
            arr.vel_flux = tsum
            arr.vel_flux_Q = fsum

    return cat


def plot_spec(freqs, spec, tstar, title=None):

    corrected_spec = spec*np.exp(2.*np.pi*freqs*tstar)

    plt.loglog(freqs, spec,  color='blue')
    plt.loglog(freqs, corrected_spec,  color='green')

    if title:
        plt.title(title)
    plt.grid()
    plt.show()

    return



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
# Filename: core.py
#  Purpose: Expansion of the obspy.core.stream module
#   Author: uquake development team
#    Email: devs@uquake.org
#
# Copyright (C) 2016 uquake development team
# --------------------------------------------------------------------
"""
Expansion of the obspy.core.stream module

:copyright:
    uquake development team (devs@uquake.org)
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""
from abc import ABC
from io import BytesIO

import numpy as np
import obspy.core.stream as obsstream
from pkg_resources import load_entry_point

from .trace import Trace
from .util import ENTRY_POINTS, tools
from .logging import logger
from pathlib import Path


class Stream(obsstream.Stream, ABC):
    __doc__ = obsstream.Stream.__doc__.replace('obspy', 'uquake')

    def __init__(self, stream=None, **kwargs):
        super(Stream, self).__init__(**kwargs)

        if stream:
            traces = []

            for tr in stream.traces:
                traces.append(Trace(trace=tr))

            self.traces = traces

    def composite(self):
        """
        returns a new stream object containing composite trace for all station.
        The amplitude of the composite traces are the norm of the amplitude of
        the trace of all component and the phase of the trace (sign) is the
        sign of the first components of a given station.
        :param st: a stream object
        :type st: ~uquake.core.stream.Stream
        :rtype: ~uquake.core.stream.Stream

        """

        return composite_traces(self)

    def select(self, **kwargs):
        if 'site' in kwargs.keys():
            trs = [tr for tr in self.traces if tr.stats.site == kwargs['site']]
        else:
            return super().select(**kwargs)

        st_tmp = Stream(traces=trs)

        kwargs_tmp = {}
        for key in kwargs.keys():
            if key == 'site':
                continue
            kwargs_tmp[key] = kwargs[key]

        return st_tmp.select(**kwargs_tmp)

    def as_array(self, wlen_sec=None, taplen=0.05):
        t0 = np.min([tr.stats.starttime for tr in self])
        sr = self[0].stats.sampling_rate

        if wlen_sec is not None:
            npts_fix = int(wlen_sec * sr)
        else:
            npts_fix = int(np.max([len(tr.data) for tr in self]))

        return tools.stream_to_array(self, t0, npts_fix, taplen=taplen), sr, t0

    def chan_groups(self):
        chanmap = self.channel_map()
        groups = [np.where(sk == chanmap)[0] for sk in np.unique(chanmap)]

        return groups

    def channel_map(self):
        stations = np.array([tr.stats.station for tr in self])
        unique = np.unique(stations)
        unique_dict = dict(zip(unique, np.arange(len(unique))))
        chanmap = np.array([unique_dict[chan] for chan in stations], dtype=int)

        return chanmap

    def write(self, filename, format='MSEED', **kwargs):

        from six import string_types
        f = filename

        if isinstance(filename, string_types):
            if filename.endswith('gz'):
                import gzip
                f = gzip.open(filename, 'w')
            elif filename.endswith('bz2'):
                import bz2
                f = bz2.BZ2File(filename, 'w')
            elif filename.endswith('zip'):
                print('Zip protocol is not supported')

        st_out = self.copy()

        return obsstream.Stream.write(st_out, f, format, **kwargs)

    write.__doc__ = obsstream.Stream.write.__doc__.replace('obspy',
                                                           'uquake')

    def write_bytes(self):
        buf = BytesIO()
        self.write(buf, format='MSEED')

        return buf.getvalue()

    def valid(self, **kwargs):
        return is_valid(self, return_stream=True)

    def concat(self, comp_st):

        c = (comp_st is not None)

        if c:
            for i, (t1, t2) in enumerate(zip(comp_st.traces, self.traces)):
                self.detrend_norm(t2)
                comp_st.traces[i] = t1.__add__(t2, method=1, fill_value=0)
        else:
            for t in self:
                self.detrend_norm(t)

            comp_st = self

        return comp_st

    @property
    def unique_stations(self):
        return np.sort(np.unique([tr.stats.station for tr in self]))

    @property
    def unique_sites(self):
        return np.sort(np.unique([tr.stats.site for tr in self]))

    @property
    def stations(self):
        return self.unique_stations

    @property
    def sites(self):
        return self.unique_sites

    def zpad_names(self):
        for tr in self.traces:
            tr.stats.station = tr.stats.station.zfill(3)
        self.sort()

    def zstrip_names(self):
        for tr in self.traces:
            tr.stats.station = tr.stats.station.lstrip('0')

    def distance_time_plot(self, event, site, scale=20, freq_min=100,
                           freq_max=1000):
        """
        plot traces that have
        :param event: event object
        :param site: site object
        :param scale: vertical size of pick markers and waveform
        :return: plot handler
        """

        st = self.copy()
        st.detrend('demean')
        st.taper(max_percentage=0.01)
        st.filter('bandpass', freqmin=freq_min, freqmax=freq_max)

        import matplotlib.pyplot as plt
        import numpy as np

        # initializing the plot

        ax = plt.subplot(111)

        if event.preferred_origin():
            origin = event.preferred_origin()
        elif event.origins:
            origin = event.origins[0]
        else:
            return

        event_location = origin.loc

        # find the earliest start time and latest end time
        start_time = None
        end_time = None

        for tr in st:
            if not start_time:
                start_time = tr.stats.starttime
                end_time = tr.stats.endtime

            if tr.stats.starttime < start_time:
                start_time = tr.stats.starttime

            if tr.stats.endtime > end_time:
                end_time = tr.stats.endtime

        for tr in st:
            station_code = tr.stats.station
            # search for arrival
            station = site.select(station_code).stations()[0]
            station_location = station.loc
            distance = np.linalg.norm(event_location - station_location)
            p_pick = None
            s_pick = None
            data = (tr.data / np.max(np.abs(tr.data))) * scale
            time_delta = tr.stats.starttime - start_time
            time = np.arange(0, len(data)) / tr.stats.sampling_rate + \
                time_delta

            for arrival in origin.arrivals:
                if arrival.get_pick().waveform_id.station_code == station_code:
                    distance = arrival.distance

                    if arrival.phase == 'P':
                        p_pick = arrival.get_pick().time - start_time
                    elif arrival.phase == 'S':
                        s_pick = arrival.get_pick().time - start_time

            ax.plot(time, data + distance, 'k')

            if p_pick:
                ax.vlines(p_pick, distance - scale, distance + scale, 'r')

            if s_pick:
                ax.vlines(s_pick, distance - scale, distance + scale, 'b')

            plt.xlabel('relative time (s)')
            plt.ylabel('distance from event (m)')

    @staticmethod
    def create_from_json_traces(traces_json_list):
        traces = []
        # for tr_json in traces_json_list:

        for i, tr_json in enumerate(traces_json_list):
            stats = tr_json['stats']
            tr = Trace.create_from_json(tr_json)
            traces.append(tr)

        return Stream(traces=traces)

    def to_traces_json(self):
        traces = []

        for tr in self:
            trout = tr.to_json()
            traces.append(trout)

        return traces

    def plot(self, *args, **kwargs):
        """
        see Obspy stream.plot()
        """
        from ..imaging.waveform import WaveformPlotting
        waveform = WaveformPlotting(stream=self, *args, **kwargs)

        return waveform.plotWaveform(*args, **kwargs)


# from uquake.core import read, read_events
# from spp.utils import application
# app = application.Application()
# site = app.get_stations()
# st = read('2018-11-08T10:21:49.898496Z.mseed', format='mseed')
# cat = read_events('test.xml')
# evt = cat[0]
# st = st.composite()


def is_valid(st_in, return_stream=False, STA=0.005, LTA=0.1, min_num_valid=5):
    """
        Determine if an event is valid or return valid traces in a  stream
        :param st_in: stream
        :type st_in: uquake.core.stream.Stream
        :param return_stream: return stream of valid traces if true else return
        true if the event is valid
        :type return_stream: bool
        :param STA: short term average used to determine if an event is valid
        :type STA: float
        :param LTA: long term average
        :type LTA: float
        :param min_num_valid: minimum number of valid traces to declare the
        event valid
        :type min_num_valid: int
        :rtype: bool or uquake.core.stream.Stream
    """

    from scipy.ndimage.filters import gaussian_filter1d
    from obspy.signal.trigger import recursive_sta_lta

    st = st_in.copy()
    st.detrend('demean').detrend('linear')
    trstd = []
    trmax = []
    trs_out = []
    st_comp = composite_traces(st)

    for tr in st_comp:
        if not np.any(tr.data):
            continue
        sampling_rate = tr.stats.sampling_rate
        trstd.append(np.std(tr.data))
        trmax.append(np.max(np.abs(tr.data)))
        nsta = int(STA * sampling_rate)
        nlta = int(LTA * sampling_rate)
        cft = recursive_sta_lta(np.array(tr.data), nsta, nlta)
        sfreq = tr.stats['sampling_rate']
        sigma = sfreq / (2 * np.pi * 100)
        cft = gaussian_filter1d(cft, sigma=sigma, mode='reflect')
        try:
            mx = np.r_[True, cft[1:] > cft[:-1]] & \
                np.r_[cft[:-1] > cft[1:], True]
        except Exception as e:
            logger.error(e)
            continue

        i1 = np.nonzero(mx)[0]
        i2 = i1[cft[i1] > np.max(cft) / 2]

        tspan = (np.max(i2) - np.min(i2)) / sampling_rate

        ratio = np.max(np.abs(tr.data)) / np.std(tr.data)

        accept = True

        if len(i2) < 3:
            if ratio < 4:
                accept = False

        elif len(i2) >= 4:
            accept = False
        # else:
        #     if ratio < 4:
        #         accept = False

        if tspan > 0.1:
            accept = False

        if (len(i2) == 2) and (tspan > 0.01) and (tspan < 0.1):
            if ratio > 5:
                accept = True

        if accept:
            for tr_accepted in st_in.select(station=tr.stats.station):
                trs_out.append(tr_accepted)

    st_out = Stream(traces=trs_out)

    if return_stream:
        return st_out
    else:
        if len(st.unique_stations()) >= min_num_valid:
            return True
        else:
            return False


def check_for_dead_trace(tr):
    eps = 1e-6
    data = tr.data.copy()
    mean = np.mean(data)
    max = np.max(data) - mean
    min = np.min(data) - mean
    # print('%s: mean:%f max:%f min:%f' % (tr.get_id(), mean, max, min))

    if max < eps and np.abs(min) < eps:
        return 1
    else:
        return 0


def composite_traces(st_in):
    """
    Requires length and sampling_rates equal for all traces
    returns a new stream object containing composite trace for all station.
    The amplitude of the composite traces are the norm of the amplitude of
    the trace of all component and the phase of the trace (sign) is the sign
    of the first components of a given station.
    :param st_in: a stream object
    :type st_in: ~uquake.core.stream.Stream
    :rtype: ~uquake.core.stream.Stream

    """

    trsout = []

    st = st_in.copy()
    st.detrend('demean')

    for site in st.unique_sites:
        trs = st.select(site=site)

        if len(trs) == 1:
            trsout.append(trs[0].copy())

            continue

        npts = len(trs[0].data)
        buf = np.zeros(npts, dtype=trs[0].data.dtype)

        for tr in trs:
            dat = tr.data
            buf += (dat - np.mean(dat)) ** 2

        buf = np.sign(trs[0].data) * np.sqrt(buf)
        stats = trs[0].stats.copy()
        ch = st_in.traces[0].stats.channel
        if len(ch) > 1:
            prefix = ch[:-1]
        stats.channel = f'{prefix}C'
        trsout.append(Trace(data=buf.copy(), header=stats))

    return Stream(traces=trsout)


def read(filename, format='MSEED', **kwargs):
    if isinstance(filename, Path):
        filename = str(filename)

    if format in ENTRY_POINTS['waveform'].keys():
        format_ep = ENTRY_POINTS['waveform'][format]
        read_format = load_entry_point(format_ep.dist.key,
                                       'obspy.plugin.waveform.%s' %
                                       format_ep.name, 'readFormat')

        st = Stream(stream=read_format(filename, **kwargs))

        # making sure the channel names are upper case
        trs = []
        for tr in st:
            tr.stats.channel = tr.stats.channel.upper()
            trs.append(tr.copy())

        st.traces = trs

        return st
    else:
        return Stream(stream=obsstream.read(filename, format=format, **kwargs))


read.__doc__ = obsstream.read.__doc__.replace('obspy', 'uquake')



import warnings
import functools


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn("Call to deprecated function {}.".format(func.__name__),
                      category=DeprecationWarning,
                      stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)
    return new_func


import copy

import matplotlib.pyplot as plt
# import numba
import numpy as np
from microquake.helpers.logging import logger
from scipy import optimize
from scipy.fftpack import rfft
from scipy.signal import savgol_filter

from microquake.core.data.inventory import get_corner_freq_from_pole, get_sensor_type_from_trace
from microquake.core.util.tools import copy_picks_to_dict


"""
    smom_measures - A collection of functions used in the calculation of
                     smom (long-period plateau) in the frequency domain
"""


def measure_pick_smom(st, inventory, event, synthetic_picks,
                      fmin=20, fmax=1000,
                      use_fixed_fmin_fmax=False,
                      plot_fit=False,
                      P_or_S='P',
                      debug_level=0):
    """
        Calculate spectra of each channel/arrival and compare to Brune model
        to fit for long_period plateau (smom)
    """

# Get P(S) spectra at all stations/channels that have a P(S) arrival:
    sta_dict = get_spectra(st, event, inventory, synthetic_picks,
                           calc_displacement=False, S_win_len=.1,
                           P_or_S=P_or_S)

    vel_dict = copy.deepcopy(sta_dict)

# Stack vel spectra to get fc ~ peak_f
# Note that fc_S is predicted to be < fc_P
    stacked_spec, freqs = stack_spectra(sta_dict)
    peak_f = peak_freq(stacked_spec, freqs, fmin=25.)

    # fit,smom_dict = calc_fit(vel_dict, fc=peak_f, fmin=fmin, fmax=fmax,
    # plot_fit=plot_fit,
    # use_fixed_fmin_fmax=use_fixed_fmin_fmax)
    # exit()

    logger.debug("pha:%s velocity stack corner freq fc=%.1f" %
                 (P_or_S, peak_f))

    if debug_level > 1:
        plot_spec(stacked_spec, freqs, title='Stack [%s] Vel spec '
                                             'peak_f=%.1f' % (P_or_S, peak_f))

# Now recalculate the spectra as Displacment spectra:

    sta_dict = get_spectra(st, event, inventory, synthetic_picks,
                           calc_displacement=True,
                           S_win_len=.1, P_or_S=P_or_S)
    stacked_spec, freqs = stack_spectra(sta_dict)
    fit, _ = calc_fit1(stacked_spec, freqs, fmin=1, fmax=fmax,
                       fit_displacement=True)

    # Calculate fmin/fmax from velocity signal/noise spec
    # and add into this displacement spec dict:

    for sta_code, sta_dd in sta_dict.items():
        for cha_code, cha_dict in sta_dd['chan_spec'].items():
            cha_dict['fmin'] = vel_dict[sta_code]['chan_spec'][cha_code][
                'fmin']
            cha_dict['fmax'] = vel_dict[sta_code]['chan_spec'][cha_code][
                'fmax']

    fit, smom_dict = calc_fit(sta_dict, fc=peak_f, fmin=fmin, fmax=fmax,
                              plot_fit=plot_fit,
                              use_fixed_fmin_fmax=use_fixed_fmin_fmax)

    phase = P_or_S
    arr_dict = {}
    arrivals = [arr for arr in event.preferred_origin().arrivals if
                arr.phase == phase]

    for arr in arrivals:
        pk = arr.get_pick()
        sta = pk.get_sta()

        if sta not in arr_dict:
            arr_dict[sta] = {}
        arr_dict[sta][phase] = arr

    for sta, sta_dict in smom_dict.items():
        smoms = []
        fits = []
        ts = []
        fmin = []
        fmax = []

        for _, cha_dict in sta_dict.items():
            smoms.append(cha_dict['smom'])
            fits.append(cha_dict['fit'])
            ts.append(cha_dict['tstar'])
            fmin.append(cha_dict['fmin'])
            fmax.append(cha_dict['fmax'])

        smom = np.sqrt(np.sum(np.array(smoms)**2))
        fit = np.sum(np.array(fits))/float(len(fits))
        fit = np.mean(fits)
        tstar = np.median(ts)
        min_f = np.max(fmin)
        max_f = np.min(fmax)

        # logger.debug("sta:%3s pha:%s smom:%12.10g ts:%.4f nchans:%d" %
        #              (sta, phase, smom, tstar, len(fits)))

        arr = arr_dict[sta][phase]
        arr.smom = smom
        arr.fit = fit
        arr.tstar = tstar
        arr.fmin = min_f
        arr.fmax = max_f
        # logger.info("sta:%3s pha:%s smom:%12.10g ts:%.4f nchans:%d fmin:%f "
        #             "fmax:%f" % (sta, phase, smom, tstar, len(fits), min_f,
        #                          max_f))

    return smom_dict, peak_f


def plot_spec(spec, freqs, title='Stacked spec'):

    plt.loglog(freqs, spec, color='blue')
    plt.xlim(1e0, 3e3)

    if title:
        plt.title(title)
    plt.grid()
    plt.show()


def peak_freq(spec_array, freqs, fmin=0., fmax=None):
    """ Return peak_f of spec_array within bounds: fmin <= freqs <= fmax

        :param spec_array: Input spectrum
        :type  spec_array: np.array
        :param freqs: Corresponding frequencies
        :type  freqs: np.array
        :param fmin: Restrict peak search above this freq
        :type  fmin: float
        :param fmax: Restrict peak search below this freq
        :type  fmax: float

        :return: peak frequency within bounds: fmin <= peak_frequency <= fmax
        :rtype: float
    """

    if fmax is None:
        fmax = np.max(freqs)

    mask = [all(f) for f in zip(freqs >= fmin, freqs <= fmax)]

    return freqs[mask][np.argmax(spec_array[mask], axis=0)]


def stack_spectra(sta_dict):
    """ stack the (normalized) pre-calculated spectra (fft) of each channel

        :param sta_dict: dictionary of pre-calculated P & S channel spectra
        (complex values) at each station
        :return: normalized_spectrum_stack (real modulus), freqs
        :rtype: (np.array, np.array)
    """

    stack = 0.
    n = 0.
    freqs = []

    for _, sta in sta_dict.items():
        for _, cha in sta['chan_spec'].items():
            if len(cha['signal_fft']) > 0:
                signal_fft = cha['signal_fft']
                # this is very dangerous assuming that the FFT will all have
                # the same frequency vector. Depending on how the FFT is
                # calculated the vector will depend on the sampling rate.
                freqs = cha['freqs']

                # Normalize each fft individually and add to stack
                stack += np.abs(signal_fft)/np.amax(np.abs(signal_fft))

                n += 1

    return stack/np.amax(stack), np.array(freqs)


def get_spectra(st, event, inventory, synthetic_picks,
                calc_displacement=False,
                S_win_len=.1, P_or_S='P'):
    """ Calculate the fft at each channel in the stream that has an arrival
    in event.arrivals

        :param st: microquake.core.stream.Stream
        :type  st: microquake.core.stream.Stream
        :param event: microquake.core.event.Event
        :type  event: microquake.core.event.Event
        :param stations: list of station metadata
        :type  stations: obspy.core.inventory.network.Network
        :param calc_displacement: If true calculate displacement spectra (
        else velocity)
        :type  calc_displacement: bool

        :return: sta_dict: dict of station P & S spectra for each channel
        :rtype: dict
    """

    pre_window_start_sec = .01
    max_S_P_time = 0.25

    # TODO Right now S win len = .1, may want to make this a function of
    #  distance and/or handle
    #     if win len exceeds trace len, etc
    origin = event.preferred_origin() if event.preferred_origin() else \
        event.origins[0]

    # Only use picks that have an arrival used for this origin:
    picks = [arr.pick_id.get_referred_object() for arr in origin.arrivals]
    pick_dict = copy_picks_to_dict(picks)

    synthetic_dict = copy_picks_to_dict(synthetic_picks)

    if P_or_S == 'P':
        arrivals = [arr for arr in origin.arrivals if arr.phase == 'P']
    else:
        arrivals = [arr for arr in origin.arrivals if arr.phase == 'S']

    arr_dict = {}

    for arr in arrivals:
        pk = arr.pick_id.get_referred_object()
        sta = pk.waveform_id.station_code
        pha = arr.phase

        if sta not in arr_dict:
            arr_dict[sta] = {}
        arr_dict[sta][pha] = arr

    arr_stations = set()

    for arr in arrivals:
        arr_stations.add(arr.pick_id.get_referred_object(
            ).waveform_id.station_code)

    st_stations = set()

    for tr in st:
        st_stations.add(tr.stats.station)

    # Only keep stations that have at least one arrival and at least one trace
    sta_codes = arr_stations.intersection(st_stations)

    # Turn into a list sorted numerically by sta code (but keep sta code = str)
    sta_codes = sorted([sta for sta in sta_codes], key=lambda x: int(x))

    # For now, let's be explicit about our assumptions:
    # For each trace, we'll use P-.01 to S-.01 to calculate the P spectrum, so
    #   let's use the max expected S-P time (at the farthest station) to fix
    #   NFFT
    dt = st[0].stats.delta  # For now, let's assume ALL traces are sampled
    # at this rate
    npts = int(max_S_P_time/dt)
    nfft = npow2(npts) * 2  # Fix nfft for all calcs
    df = 1/(float(nfft)*dt)
    # print("%s: Set nfft=%d --> df=%.3f" % (fname, nfft, df))

    # 1. Create a dict of keys=sta_code for all 'P' arrivals with the
    # necessary pick/sta info inside
    sta_dict = {}

    for sta_code in sta_codes:

        d = {}

        if 'P' in pick_dict[sta_code] and pick_dict[sta_code][
            'P'].evaluation_status != "rejected":
            d['ptime'] = pick_dict[sta_code]['P'].time
        else:
            d['ptime'] = synthetic_dict[sta_code]['P'].time

        if 'S' in pick_dict[sta_code] and pick_dict[sta_code][
            'S'].evaluation_status != "rejected":
            d['stime'] = pick_dict[sta_code]['S'].time
        else:
            d['stime'] = synthetic_dict[sta_code]['S'].time

        sta_loc = inventory.select(sta_code).loc

        if sta_loc is None:
            raise Exception(f"Oops: sta:{sta_code} not found in inventory!")

        d['loc'] = sta_loc

        # Try to use distance along ray path if set in the arrival dict,
        #   else use Euclidean distance
        dist = arr_dict[sta_code][P_or_S].get('hypo_dist_in_m', None)

        if dist is None:
            dist = np.linalg.norm(d['loc'] - origin.loc)  # Dist in meters

        d['R'] = dist
        sta_dict[sta_code] = d


# 2. Calc/save signal/noise fft spectra at all channels that have P(S)
    # arrivals:
    st.detrend('demean').detrend('linear').taper(type='cosine',
                                                 max_percentage=0.05,
                                                 side='both')

    for sta_code, sta in sta_dict.items():

        if P_or_S == 'P':
            signal_start = sta['ptime'] - pre_window_start_sec
            signal_end = sta['stime'] - pre_window_start_sec
        else:
            signal_start = sta['stime'] - pre_window_start_sec
            signal_end = sta['stime'] + S_win_len

        signal_len = signal_end - signal_start

        noise_end = sta['ptime'] - pre_window_start_sec
        noise_start = noise_end - signal_len

        if signal_start > signal_end:
            logger.info("Signal start time larger the end time, continuing")
            logger.info("*** sta:%3s ptime:%s stime:%s P_or_S:%s pre_win:%f "
                        "S_win:%f" % (sta_code, sta['ptime'], sta['stime'],
                                      P_or_S, pre_window_start_sec, S_win_len))

            continue

        trs = st.select(station=sta_code)

        if trs:
            chans = {}

            for tr in trs:
                ch = {}

                cha_code = tr.stats.channel

                signal = tr.copy()
                signal.trim(starttime=signal_start, endtime=signal_end)
                signal.detrend('demean').detrend('linear')
                signal.taper(type='cosine', max_percentage=0.05, side='both')

                noise = tr.copy()
            # if noise_start < trace.stats.starttime - then what ?
                noise.trim(starttime=noise_start, endtime=noise_end)
                noise.detrend('demean').detrend('linear')
                noise.taper(type='cosine', max_percentage=0.05, side='both')

                sensor_type = get_sensor_type_from_trace(tr)

                if sensor_type == 'ACC':
                    signal.integrate().taper(type='cosine',
                                             max_percentage=0.05, side='both')
                    noise.integrate().taper(type='cosine',
                                            max_percentage=0.05, side='both')
                elif sensor_type == 'VEL':
                    pass
                elif sensor_type == 'DISP':
                    logger.info("Not yet set up to handle input traces = "
                                "DISPLACMENT !!")

                    continue
                else:
                    logger.error("sta:%3s cha:%s sensor_type=[%s] is unknown" %
                                 (sta_code, cha_code, sensor_type))

                    continue

                if calc_displacement:
                    signal.integrate().taper(type='cosine',
                                             max_percentage=0.05, side='both')
                    noise.integrate().taper(type='cosine',
                                            max_percentage=0.05, side='both')

                # check(signal)
                # parsevals(signal.data, dt, nfft)

                (signal_fft, freqs) = unpack_rfft(rfft(signal.data, n=nfft), df)
                (noise_fft, freqs) = unpack_rfft(rfft(noise.data, n=nfft), df)

                # MTH: Determine the valid freq range: fmin - fmax
                #      To fit to model.
                # fmin/fmax = where smoothed signal spec exceeds smoothed
                # noise spec by snr_threshold
                # fc1 = low-frequency corner of this trace's sensor

                if not calc_displacement:

                    fc1 = get_corner_freq_from_pole(
                        tr.stats.response.get_paz().poles[0])
                    fmin, fmax = find_fmin_fmax_from_spec(signal_fft,
                                                          noise_fft, 1.3,
                                                          200., df)

                    # Take the maximum of fc1 and fmin to use for fitting:
                    # fmin is None if signal > noise all the way to DC

                    if fmin is None or fc1 > fmin:
                        fmin = fc1

                    if fmax is None:
                        fmax = freqs[-1]

                    ch['fmin'] = fmin
                    ch['fmax'] = fmax

                # plot_spec(freqs, signal_fft, noise_fft, title=None)

                # np/scipy ffts are not scaled by dt
                signal_fft *= dt
                noise_fft *= dt
                '''
                This is a 1-sided FFT (ie, we only have the +ve frequency half)
                To conserve power a la Parseval's, we need to double
                the power at each frequency *except* for DC & Nyquist
                which are purely real and don't have a -ve freq.

                Equivalently, we scale the FFT amps at these frequencies
                by sqrt(2):
                '''
                signal_fft[1:-1] *= np.sqrt(2.)
                noise_fft[1:-1] *= np.sqrt(2.)

                # plot_signal(signal, noise)

                ch['nfft'] = nfft
                ch['dt'] = dt
                ch['df'] = df
                ch['freqs'] = freqs
                ch['signal_fft'] = signal_fft
                ch['noise_fft'] = noise_fft
                ch['P_or_S'] = P_or_S

                if calc_displacement:
                    ch['disp_or_vel'] = 'DISPLACEMENT'
                else:
                    ch['disp_or_vel'] = 'VELOCITY'
                chans[tr.stats.channel] = ch
            sta['chan_spec'] = chans

            # end for tr in trs
        # end if trs
    # end for sta in sta_dict

    return sta_dict


def brune_omega3(fc: float, mom: float, ts: float, f: float) -> float:
    return mom * 1. / (1 + np.power(f/fc, 3.)) * np.exp(-np.pi * ts * f)


# @numba.jit(nopython=True)
def brune_dis_spec(fc: float, mom: float, ts: float, f: float) -> float:
    return mom * (fc * fc) / (fc * fc + f * f) * np.exp(-np.pi * ts * f)


def brune_vel_spec(fc: float, mom: float, ts: float, f: float) -> float:
    return 2. * np.pi * f * brune_dis_spec(fc, mom, ts, f)


def brune_acc_spec(fc: float, mom: float, ts: float, f: float) -> float:
    return 2. * np.pi * f * brune_vel_spec(fc, mom, ts, f)


def getresidfit(data_spec, model_func, fc: float, freqs: list, Lnorm='L1',
                weight_fr=False, fmin=1., fmax=3000.) -> float:

    # Give it a starting vector:
    #     smom    tstar
    # MTH: this should be wert origin time anyway ...
    # tt_s = sta['stime'] - tr.stats.starttime
    # pp = [1e-5, tt_s/500.]
    # pp = [1e-5, tt_s/200.]
    tt_s = .02
    start = np.array([1e-10, tt_s/200.])

    # @numba.jit(nopython=True)
    def inner(p):
        smom = p[0]
        ts = p[1]
        fit = 0.
        # Neither Powell nor Nelder-Meade seem to have a way to restrict
        # search params to be > 0,
        # and L-BFGS-B does not appear to work, so here's a hack:

        if smom < 0. or ts < 0.:
            return 1e12

        for i, f in enumerate(freqs):
            if f < fmin or f > fmax:
                continue

            res = model_func(fc, smom, ts, f)

            if res < 0.:
                # logger.info("** OOPS: f=%f smom=%g ts=%g model_func=%g" %
                # (f, smom, ts, res))
                pass
            diff = np.log10(data_spec[i]) - np.log10(res)

            if Lnorm == 'L2':
                diff *= diff

            if weight_fr:
                diff /= f
            fit += diff

        return fit

    (sol, fopt, *rest) = optimize.fmin(inner, start, ftol=.01, disp=False,
                                       full_output=1)

    return (sol, fopt)


def getresidfit2(data_spec, model_func, freqs: list, Lnorm='L1',
                 weight_fr=False, fmin=1., fmax=3000.) -> float:

    def inner(p):
        smom = p[0]
        fc = p[1]
        fit = 0.
        # Neither Powell nor Nelder-Meade seem to have a way to restrict
        # search params to be > 0,
        # and L-BFGS-B does not appear to work, so here's a hack:

        if smom < 0.:
            return 1e12

        for i, f in enumerate(freqs):
            if f < fmin or f > fmax:
                continue
            diff = np.log10(data_spec[i]) - np.log10(model_func(fc, smom, f))

            if Lnorm == 'L2':
                diff *= diff

            if weight_fr:
                diff /= f
            fit += diff

        return fit

    return inner


def getresidfit3(data_spec, model_func, freqs: list, Lnorm='L1',
                 weight_fr=False, fmin=1., fmax=3000.) -> float:

    def inner(p):
        smom = p[0]
        ts = p[1]
        fc = p[2]
        fit = 0.
        # Neither Powell nor Nelder-Meade seem to have a way to restrict
        # search params to be > 0,
        # and L-BFGS-B does not appear to work, so here's a hack:

        if smom < 0. or ts < 0.:
            return 1e12

        for i, f in enumerate(freqs):
            if f < fmin or f > fmax:
                continue

            res = model_func(fc, smom, ts, f)

            if res < 0.:
                logger.info("** OOPS: f=%f smom=%g ts=%g model_func=%g" % (
                    f,  smom, ts, res))
                pass
            diff = np.log10(data_spec[i]) - np.log10(res)

            if Lnorm == 'L2':
                diff *= diff

            if weight_fr:
                diff /= f
            fit += diff

        return fit

    return inner


def calc_fit1(spec, freqs, fmin=1., fmax=1000., Lnorm='L2', weight_fr=False,
              fit_displacement=False):
    # spec = spec_amp

    model_func = brune_vel_spec

    if fit_displacement:
        model_func = brune_dis_spec

    residfit = getresidfit3(spec, model_func, freqs, Lnorm=Lnorm,
                            weight_fr=weight_fr, fmin=fmin, fmax=fmax)

    # Give it a starting vector:
    #    smom  t* fc
    pp = [1., .001, 100]
    (sol, fit, *rest) = optimize.fmin(residfit, np.array(pp), xtol=10**-12,
                                      ftol=10**-6, disp=False, full_output=1)
    mom, ts, fc = sol[0], sol[1], sol[2]
    # print(fc)

    if sol[0] < 0.:
        logger.info("Ummm smom < 0 !!!")

    # print("solution=",sol)
    plot_fit = 0

    if plot_fit:
        model_spec = np.array(np.zeros(freqs.size), dtype=np.float_)

        for i, f in enumerate(freqs):
            model_spec[i] = model_func(fc, mom, ts, f)
        plot_spec2(freqs, spec, model_spec, title="Fit to spec stack "
                                                  "fc=%.2f" % fc)

    return fit, fc


def calc_fit(sta_dict, fc, fmin=20., fmax=1000.,
             Lnorm='L2', weight_fr=False,
             plot_fit=False,
             use_fixed_fmin_fmax=False):

    # frequencies = []
    fit = 0.
    smom_dict = {}

    # for _, sta in sta_dict.items():
    #     if 'chan_spec' not in sta:
    #         logger.debug("sta:%3s: Has no chan_spec --> Skip" % sta_code)

    #         continue
    #     dd = {}

    #     for _, cha in sta['chan_spec'].items():
    #         freqs = cha['freqs']
    #         frequencies = set(list(frequencies) + list(freqs))

    # for f in frequencies:
    #     res_dict[(fc, f)] = (fc * fc) / (fc * fc + f * f)

    # (sol,fopt, _)= optimize.fmin(residfit, np.array(pp),xtol=10**-12,
    # ftol=10**-12,disp=False, full_output=1)

    for sta_code, sta in sta_dict.items():
        if 'chan_spec' not in sta:
            logger.debug("sta:%3s: Has no chan_spec --> Skip" % sta_code)

            continue
        dd = {}

        for cha_code, cha in sta['chan_spec'].items():
            signal_fft, freqs, noise_fft = (cha['signal_fft'], cha['freqs'],
                                            cha['noise_fft'])

            model_func = brune_vel_spec

            if cha['disp_or_vel'] == 'DISPLACEMENT':
                model_func = brune_dis_spec

            ch_dict = {}
            # Optimize fit of model_func to signal_fft to solve for smom:

            if not use_fixed_fmin_fmax:
                fmin = cha['fmin']
                fmax = cha['fmax']

            # logger.debug("calc_fit sta:%3s cha:%s fmin:%.1f fmax:%.1f" %
            #              (sta_code, cha_code, fmin, fmax))

            (sol, fopt) = getresidfit(np.abs(signal_fft), model_func, fc,
                                      freqs, Lnorm=Lnorm,
                                      weight_fr=weight_fr, fmin=fmin,
                                      fmax=fmax)

            if sol[0] < 0.:
                logger.info("Ummm smom < 0 !!!")

            # gives same result as fopt
            # ch_fit = residfit(sol)
            fit += fopt

            ch_dict['smom'] = sol[0]
            ch_dict['tstar'] = sol[1]
            ch_dict['fit'] = fopt
            ch_dict['P_or_S'] = cha['P_or_S']
            ch_dict['fmin'] = cha['fmin']
            ch_dict['fmax'] = cha['fmax']

            # logger.debug("calc_fit: sta:%s cha:%s smom:%12.10g fit:%.2f
            # ts=%f" %
            #              (sta_code, cha_code, sol[0], fopt, sol[1]))

            # if plot_fit and cha['P_or_S'] == "S":
            plot_fit = 0

            if plot_fit:
                model_spec = np.array(np.zeros(freqs.size), dtype=np.float_)
                w2_no_atten = np.array(np.zeros(freqs.size), dtype=np.float_)
                w3_no_atten = np.array(np.zeros(freqs.size), dtype=np.float_)
                brune_vel = np.array(np.zeros(freqs.size), dtype=np.float_)
                corrected_vel_spec = np.array(np.zeros(freqs.size),
                                              dtype=np.float_)
                Q = 100
                tstar = sta['R']/(5.3e3 * Q)
                logger.debug("tstar=", tstar)

                for i, f in enumerate(freqs):
                    # print("i=%d f=%f" % (i, f))
                    model_spec[i] = model_func(fc, sol[0], sol[1], f)
                    w2_no_atten[i] = model_func(fc, sol[0], 0, f)
                    w3_no_atten[i] = brune_omega3(fc, sol[0], 0, f)
                    # brune_vel[i] = brune_vel_spec(fc, sol[0], sol[1], f)
                    brune_vel[i] = brune_vel_spec(fc, sol[0], 0, f)
                    corrected_vel_spec[i] = signal_fft[i] * np.exp(
                        np.pi*f*tstar)

                # plot_spec5(freqs, vel_spec,  np.abs(noise_fft), model_spec,
                # plot_spec5(freqs, np.abs(signal_fft),  np.abs(noise_fft),
                # model_spec,
                plot_spec5(freqs, np.abs(corrected_vel_spec),  np.abs(
                    noise_fft), model_spec,
                           w2_no_atten, brune_vel,
                           title="sta:%s ch:%s spec fit phase:%s ["
                                 "lambda=%12.10g]" %
                           (sta_code, cha_code, ch_dict['P_or_S'], sol[0]),
                           subtitle="fmin:%.1f fmax:%.1f "
                                    "use_fixed_fmin_fmax:%s" %
                           (fmin, fmax, use_fixed_fmin_fmax))

            dd[cha_code] = ch_dict
        smom_dict[sta_code] = dd

    # logger.info(len(ldict))

    return fit, smom_dict


def plot_spec5(freqs, signal_fft, noise_fft, model_spec,
               model2_spec, model3_spec,
               title=None, subtitle=None):

    # plt.loglog(freqs, np.abs(signal_fft), color='blue')
    plt.loglog(freqs, np.abs(signal_fft), color='blue', linestyle="--")
    # plt.loglog(freqs, np.abs(noise_fft),  color='red')
    plt.loglog(freqs, model2_spec,  color='black')
    plt.loglog(freqs, model_spec,  color='green')
    # plt.loglog(freqs, model_spec,  color='green', linewidth=.5)
    plt.loglog(freqs, model3_spec,  color='red')

    plt.legend(['signal', 'w^2', 'w^2 + ts', 'Brune vel'])
    # plt.legend(['signal', 'w^2', 'w^2 + ts', 'w^3'])
    # plt.legend(['signal', 'noise', 'model'])

    plt.xlim(1e-3, 3e3)
    # plt.ylim(1e-12, 1e-6)
    # plt.ylim(1e-15, 1e-9)
    # suptitle is larger and goes on top (?!)

    if title:
        plt.suptitle(title)

    if subtitle:
        plt.title(subtitle)
    plt.grid()
    plt.show()


def plot_spec3(freqs, signal_fft, noise_fft, model_spec, title=None,
               subtitle=None):

    plt.loglog(freqs, np.abs(signal_fft), color='blue')
    plt.loglog(freqs, np.abs(noise_fft),  color='red')
    plt.loglog(freqs, model_spec,  color='green')
    plt.legend(['signal', 'noise', 'model'])
    plt.xlim(1e0, 3e3)
    # plt.ylim(1e-12, 1e-6)
    plt.ylim(1e-15, 1e-9)
    # suptitle is larger and goes on top (?!)

    if title:
        plt.suptitle(title)

    if subtitle:
        plt.title(subtitle)
    plt.grid()
    plt.show()


def find_fmin_fmax_from_spec(signal_fft, noise_fft, snr_thresh, fc, df):

    npoly = 2
    nsmooth = 31
    signal = savgol_filter(np.abs(signal_fft), nsmooth, npoly)
    noise = savgol_filter(np.abs(noise_fft), nsmooth, npoly)
    snr = signal/noise

    fmin = None
    fmax = None

    ic = int(fc/df)
    i1 = -999
    i2 = -999

    for i in range(ic, 0, -1):
        if snr[i] <= snr_thresh:
            i1 = i

            break

    if i1 > 0:
        fmin = float(i1)*df

    for i in range(ic, snr.size):
        if snr[i] <= snr_thresh:
            i2 = i

            break

    if i2 > 0:
        fmax = float(i2)*df

    return fmin, fmax


def plot_spec(freqs, signal_fft, noise_fft=None, title=None):

    npoly = 2
    nsmooth = 31
    signal = savgol_filter(np.abs(signal_fft), nsmooth, npoly)
    noise = savgol_filter(np.abs(noise_fft), nsmooth, npoly)
    # signal = np.abs(signal_fft)
    # noise = np.abs(noise_fft)

    # plt.loglog(freqs, np.abs(signal_fft), color='blue')
    plt.loglog(freqs, signal, color='blue')

    if noise_fft is not None:
        plt.loglog(freqs, noise,  color='red')
        # plt.loglog(freqs, np.abs(noise_fft),  color='red')
        plt.legend(['signal', 'noise'])
    # plt.loglog(freqs, model_spec,  color='green')
    # plt.legend(['signal', 'noise', 'model'])
    plt.xlim(1e0, 3e3)
    # plt.ylim(1e-8, 1.2e-4)
    # plt.ylim(1e-12, 1e-4)

    if title:
        plt.title(title)
    plt.grid()
    plt.show()


def plot_spec2(freqs, spec, model_spec, title=None):

    plt.loglog(freqs, spec, color='blue')
    plt.loglog(freqs, model_spec,  color='green')
    plt.legend(['signal', 'model'])
    plt.xlim(1e0, 3e3)
    # plt.ylim(1e-12, 1e-4)

    if title:
        plt.title(title)
    plt.grid()
    plt.show()


def plot_signal(signal, noise=None):
    # def plot_signal(signal, noise):
    '''
        signal = obspy Trace windowed around signal
        noise  = obspy Trace windowed around pre-signal noise
    '''
    t = []
    dt = signal.stats.sampling_rate

    for i in range(signal.stats.npts):
        t.append(float(i)*dt)
    plt.plot(t, signal, color='blue')

    if noise:
        plt.plot(t, noise, color='red')
        plt.legend(['signal', 'noise'])
    plt.title("%s: P window" % signal.get_id())
    plt.grid()
    plt.show()


# @numba.jit(nopython=True)
def unpack_rfft(rfft, df):
    n = rfft.size

    if n % 2 == 0:
        n2 = int(n/2)
    else:
        # n2 = int((n + 1) / 2)
        raise Exception("n is odd!!")
    # print("n2=%d" % n2)

    c_arr = np.zeros(n2+1, dtype=np.complex_)
    freqs = np.zeros(n2+1, dtype=np.float_)

    c_arr[0] = rfft[0]
    c_arr[n2] = rfft[n-1]
    freqs[0] = 0.
    freqs[n2] = float(n2)*df

    for i in range(1, n2):
        freqs[i] = float(i)*df
        c_arr[i] = np.complex(rfft[2*i - 1], rfft[2*i])

    return c_arr, freqs


def smooth(x, window_len=11, window='hanning'):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    input:
        x: the input signal
        window_len: the dimension of the smoothing window; should be an odd
        integer
        window: the type of window from 'flat', 'hanning', 'hamming',
        'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array
    instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(
    window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")
        # raise ValueError, "smooth only accepts 1 dimension arrays."

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")
        # raise ValueError, "Input vector needs to be bigger than window size."

    if window_len < 3:
        return x

    if window not in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', "
                         "'bartlett', 'blackman'")
        # raise ValueError, "Window is on of 'flat', 'hanning', 'hamming',
        # 'bartlett', 'blackman'"

    s = np.r_[x[window_len-1:0:-1], x, x[-2:-window_len-1:-1]]
    # print(len(s))

    if window == 'flat':  # moving average
        w = np.ones(window_len, 'd')
    else:
        w = eval('np.'+window+'(window_len)')

    y = np.convolve(w/w.sum(), s, mode='valid')

    return y


def npow2(n: int) -> int:
    """ return power of 2 >= n
    """

    if n <= 0:
        return 0

    nfft = 2

    while (nfft < n):
        nfft = (nfft << 1)

    return nfft



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
# Filename: <filename>
#  Purpose: <purpose>
#   Author: <author>
#    Email: <email>
#
# Copyright (C) <copyright>
# --------------------------------------------------------------------
"""


:copyright:
    <copyright>
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""
import numpy as np
from .base import Grid
from pathlib import Path
from uuid import uuid4
import matplotlib.pyplot as plt
from loguru import logger
import skfmm
from multiprocessing import Pool, cpu_count
from functools import partial
from typing import Optional
import h5py
from .base import ray_tracer


__cpu_count__ = cpu_count()


valid_phases = ('P', 'S')

valid_grid_types = (
    'VELOCITY',
    'VELOCITY_METERS',
    'SLOWNESS',
    'VEL2',
    'SLOW2',
    'SLOW2_METERS',
    'SLOW_LEN',
    'STACK',
    'TIME',
    'TIME2D',
    'PROB_DENSITY',
    'MISFIT',
    'ANGLE',
    'ANGLE2D'
)

valid_float_types = {
    # NLL_type: numpy_type
    'FLOAT': 'float32',
    'DOUBLE': 'float64'
}

valid_grid_units = (
    'METER',
    'KILOMETER',
)

__velocity_grid_location__ = Path('model')
__time_grid_location__ = Path('time')

__default_grid_units__ = 'METER'
__default_float_type__ = 'FLOAT'


def validate_phase(phase):
    if phase not in valid_phases:
        msg = f'phase should be one of the following valid phases:\n'
        for valid_phase in valid_phases:
            msg += f'{valid_phase}\n'
        raise ValueError(msg)
    return True


def validate_grid_type(grid_type):
    if grid_type.upper() not in valid_grid_types:
        msg = f'grid_type = {grid_type} is not valid\n' \
              f'grid_type should be one of the following valid grid ' \
              f'types:\n'
        for valid_grid_type in valid_grid_types:
            msg += f'{valid_grid_type}\n'
        raise ValueError(msg)
    return True


def validate_grid_units(grid_units):
    if grid_units.upper() not in valid_grid_units:
        msg = f'grid_units = {grid_units} is not valid\n' \
              f'grid_units should be one of the following valid grid ' \
              f'units:\n'
        for valid_grid_unit in valid_grid_units:
            msg += f'{valid_grid_unit}\n'
        raise ValueError(msg)
    return True


def validate_float_type(float_type):
    if float_type.upper() not in valid_float_types.keys():
        msg = f'float_type = {float_type} is not valid\n' \
              f'float_type should be one of the following valid float ' \
              f'types:\n'
        for valid_float_type in valid_float_types:
            msg += f'{valid_float_type}\n'
        raise ValueError(msg)
    return True


def validate(value, choices):
    if value not in choices:
        msg = f'value should be one of the following choices\n:'
        for choice in choices:
            msg += f'{choice}\n'
        raise ValueError(msg)
    return True


def read_grid(base_name, path='.', float_type=__default_float_type__):
    """
    read two parts NLLoc files
    :param base_name:
    :param path: location of grid files
    :param float_type: float type as defined in NLLoc grid documentation
    """
    header_file = Path(path) / f'{base_name}.hdr'
    with open(header_file, 'r') as in_file:
        line = in_file.readline()
        line = line.split()
        shape = tuple([int(line[0]), int(line[1]), int(line[2])])
        origin = np.array([float(line[3]), float(line[4]),
                           float(line[5])]) * 1000
        spacing = np.array([float(line[6]), float(line[7]),
                            float(line[8])]) * 1000

        grid_type = line[9]
        grid_unit = 'METER'

        line = in_file.readline()

        if grid_type in ['ANGLE', 'ANGLE2D', 'TIME', 'TIME2D']:
            line = line.split()
            seed_label = line[0]
            seed = (float(line[1]) * 1000,
                    float(line[2]) * 1000,
                    float(line[3]) * 1000)

        else:
            seed_label = None
            seed = None

    buf_file = Path(path) / f'{base_name}.buf'
    if float_type == 'FLOAT':
        data = np.fromfile(buf_file,
                           dtype=np.float32)
    elif float_type == 'DOUBLE':
        data = np.fromfile(buf_file,
                           dtype=np.float64)
    else:
        msg = f'float_type = {float_type} is not valid\n' \
              f'float_type should be one of the following valid float ' \
              f'types:\n'
        for valid_float_type in valid_float_types:
            msg += f'{valid_float_type}\n'
        raise ValueError(msg)

    data = data.reshape(shape)

    if '.P.' in base_name:
        phase = 'P'
    else:
        phase = 'S'

    # reading the model id file
    mid_file = Path(path) / f'{base_name}.mid'
    if mid_file.exists():
        with open(mid_file, 'r') as mf:
            model_id = mf.readline().strip()

    else:
        model_id = str(uuid4())

        # (self, base_name, data_or_dims, origin, spacing, phase,
        #  seed=None, seed_label=None, value=0,
        #  grid_type='VELOCITY_METERS', grid_units='METER',
        #  float_type="FLOAT", model_id=None):

    network_code = base_name.split('.')[0]
    if grid_type in ['VELOCITY', 'VELOCITY_METERS']:
        return VelocityGrid3D(network_code, data, origin, spacing, phase=phase,
                              model_id=model_id)

    elif grid_type == 'TIME':
        return TTGrid(network_code, data, origin, spacing, seed,
                      seed_label, phase=phase, model_id=model_id)

    elif grid_type == 'ANGLE':
        return AngleGrid(network_code, data, origin, spacing, seed,
                         seed_label, angle_type='AZIMUTH', phase=phase,
                         model_id=model_id)

    else:
        return NLLocGrid(data, origin, spacing, phase,
                         grid_type=grid_type, model_id=model_id,
                         grid_units=grid_unit)


class Seeds:
    __valid_measurement_units__ = ['METERS', 'KILOMETERS']

    def __init__(self, sites=[], units='METERS'):
        """
        specifies a series of source location from an inventory object
        :param sites: a list of sites containing at least the location,
        and site label
        :type sites: list of dictionary

        :Example:

        >>> site = {'label': 'test', 'x': 1000, 'y': 1000, 'z': 1000,
                      'elev': 0.0}
        >>> sites = [site]
        >>> seeds = Seeds(sites)

        """

        validate(units, self.__valid_measurement_units__)
        self.units = units

        self.sites = sites

    @classmethod
    def from_inventory(cls, inventory):
        """
        create from an inventory object
        :param inventory:
        :type inventory: uquake.core.inventory.Inventory
        """

        srces = []
        for site in inventory.sites:
            srce = {'label': site.code,
                    'x': site.x,
                    'y': site.y,
                    'z': site.z,
                    'elev': 0}
            srces.append(srce)

        return cls(srces)

    @classmethod
    def from_json(cls, json):
        pass

    def add(self, label, x, y, z, elev=0, units='METERS'):
        """
        Add a single site to the source list
        :param label: site label
        :type label: str
        :param x: x location relative to geographic origin expressed
        in the units of measurements for site/source
        :type x: float
        :param y: y location relative to geographic origin expressed
        in the units of measurements for site/source
        :type y: float
        :param z: z location relative to geographic origin expressed
        in the units of measurements for site/source
        :type z: float
        :param elev: elevation above z grid position (positive UP) in
        kilometers for site (Default = 0)
        :type elev: float
        :param units: units of measurement used to express x, y, and z
        ( 'METERS' or 'KILOMETERS')

        """

        validate(units.upper(), self.__valid_measurement_units__)

        self.sites.append({'label': label, 'x': x, 'y': y, 'z': z,
                           'elev': elev})

        self.units = units.upper()

    @staticmethod
    def generate_random_seeds_in_grid(grid, nb_seeds=1):
        """
        generate nb_seeds random seeds inside the grid provided. This function
        is mainly used for testing purposes
        :param grid: a grid
        :type grid: uquake.grid.base.Grid or an object inheriting from Grid
        :param nb_seeds: number of seeds to generate
        :return: a list of seeds

        >>> from uquake.grid.base import Grid
        >>> from uquake.grid.nlloc import Seeds
        >>> grid_dimensions = [10, 10, 10]
        >>> grid_spacing = [1, 1, 1]
        >>> grid_origin = [0, 0, 0]
        >>> grid = Grid(grid_dimensions, grid_spacing, grid_origin, value=1)
        >>> seeds = Seeds.generate_random_seeds_in_grid(grid, nb_seeds=10)
        """

        seeds = Seeds()
        label_root = 'seed'
        for i, point in enumerate(grid.generate_random_points_in_grid(
                nb_points=nb_seeds)):
            label = f'{label_root}_{i}'
            seeds.add(label, point[0], point[1], point[2])

        return seeds

    def __repr__(self):
        line = ""

        for site in self.sites:

            # test if site name is shorter than 6 characters

            line += f'GTSRCE {site["label"]} XYZ ' \
                    f'{site["x"] / 1000:>15.6f} ' \
                    f'{site["y"] / 1000:>15.6f} ' \
                    f'{site["z"] / 1000:>15.6f} ' \
                    f'0.00\n'

        return line

    @property
    def locs(self):
        seeds = []
        for site in self.sites:
            seeds.append([site['x'], site['y'], site['z']])
        return np.array(seeds)

    @property
    def labels(self):
        seed_labels = []
        for site in self.sites:
            seed_labels.append(site['label'])

        return np.array(seed_labels)


class Srces(Seeds):
    def __init__(self, sites=[], units='METERS'):
        super().__init__(sites=sites, units=units)


class NLLocGrid(Grid):
    """
    base 3D rectilinear grid object
    """
    def __init__(self, data_or_dims, origin, spacing, phase,
                 value=0, grid_type='VELOCITY_METERS',
                 grid_units=__default_grid_units__,
                 float_type="FLOAT", model_id=None):
        """
        :param data_or_dims: data or data dimensions. If dimensions are
        provided the a homogeneous gris is created with value=value
        :param origin: origin of the grid
        :type origin: list
        :param spacing: the spacing between grid nodes
        :type spacing: list
        :param phase: the useis phase (value 'P' or 'S')
        :type phase: str
        :param value:
        :type value: float
        :param grid_type:
        :type grid_type: str
        :param grid_units:
        :type grid_units: str
        :param float_type:
        :type float_type: str
        :param model_id:
        :type model_id: str
        """

        super().__init__(data_or_dims, spacing=spacing, origin=origin,
                         value=value, resource_id=model_id)

        if validate_phase(phase):
            self.phase = phase.upper()

        if validate_grid_type(grid_type):
            self.grid_type = grid_type.upper()

        # if grid_type.upper() in ['TIME', 'TIME2D', 'ANGLE', 'ANGLE2D']:
        #     if not seed:
        #         raise ValueError('the seeds value must be set for TIME and '
        #                          'ANGLE grids')
        #     if not seed_label:
        #         raise ValueError('the seed_label must be set for TIME '
        #                          'and ANGLE grids')

        # self.seed = seed
        # self.seed_label = seed_label

        if validate_grid_units(grid_units):
            self.grid_units = grid_units.upper()

        if validate_float_type(float_type):
            self.float_type = float_type.upper()

    def _write_grid_data(self, base_name, path='.'):

        Path(path).mkdir(parents=True, exist_ok=True)

        with open(Path(path) / (base_name + '.buf'), 'wb') \
                as out_file:
            if self.float_type == 'FLOAT':
                out_file.write(self.data.astype(np.float32).tobytes())

            elif self.float_type == 'DOUBLE':
                out_file.write(self.data.astype(np.float64).tobytes())

    def _write_grid_header(self, base_name, path='.', seed_label=None,
                           seed=None, seed_units=None):

        # convert 'METER' to 'KILOMETER'
        if self.grid_units == 'METER':
            origin = self.origin / 1000
            spacing = self.spacing / 1000
        else:
            origin = self.origin
            spacing = self.spacing

        line1 = f'{self.shape[0]:d} {self.shape[1]:d} {self.shape[2]:d}  ' \
                f'{origin[0]:f} {origin[1]:f} {origin[2]:f}  ' \
                f'{spacing[0]:f} {spacing[1]:f} {spacing[2]:f}  ' \
                f'{self.grid_type}\n'

        with open(Path(path) / (base_name + '.hdr'), 'w') as out_file:
            out_file.write(line1)

            if self.grid_type in ['TIME', 'ANGLE']:

                if seed_units is None:
                    logger.warning(f'seed_units are not defined. '
                                   f'Assuming same units as grid ('
                                   f'{self.grid_units}')
                if self.grid_units == 'METER':
                    seed = seed / 1000

                line2 = u"%s %f %f %f\n" % (seed_label,
                                            seed[0], seed[1], seed[2])
                out_file.write(line2)

            out_file.write(u'TRANSFORM  NONE\n')

        return True

    def _write_grid_model_id(self, base_name, path='.'):
        with open(Path(path) / (base_name + '.mid'), 'w') as out_file:
            out_file.write(f'{self.model_id}')
        return True

    def write(self, base_name, path='.'):

        self._write_grid_data(base_name, path=path)
        self._write_grid_header(base_name, path=path)
        self._write_grid_model_id(base_name, path=path)

        return True

    @property
    def model_id(self):
        return self.resource_id

    @property
    def site(self):
        return self.seed_label


class ModelLayer:
    """
    1D model varying in Z
    """

    def __init__(self, z_top, value_top):
        """
        :param z_top: Top of the layer z coordinates
        :param value_top: Value at the top of the layer
        """
        self.z_top = z_top
        self.value_top = value_top

    def __repr__(self):
        return f'top - {self.z_top:5d} | value - {self.value_top:5d}\n'


class LayeredVelocityModel(object):

    def __init__(self, model_id=None, velocity_model_layers=None,
                 phase='P', grid_units='METER',
                 float_type=__default_float_type__):
        """
        Initialize
        :param model_id: model id, if not set the model ID is set using UUID
        :type model_id: str
        :param velocity_model_layers: a list of VelocityModelLayer
        :type velocity_model_layers: list
        :param phase: Phase either 'P' or 'S'
        :type phase: str
        """

        if velocity_model_layers is None:
            self.velocity_model_layers = []

        if validate_phase(phase):
            self.phase = phase.upper()

        if validate_grid_units(grid_units):
            self.grid_units = grid_units.upper()

        if validate_float_type(float_type):
            self.float_type = float_type.upper()

        self.grid_type = 'VELOCITY'

        if model_id is None:
            model_id = str(uuid4())

        self.model_id = model_id

    def __repr__(self):
        output = ''
        for i, layer in enumerate(self.velocity_model_layers):
            output += f'layer {i+1:4d} | {layer}'

        return output

    def add_layer(self, layer):
        """
        Add a layer to the model. The layers must be added in sequence from the
        top to the bottom
        :param layer: a LayeredModel object
        """
        if not (type(layer) is ModelLayer):
            raise TypeError('layer must be a VelocityModelLayer object')

        if self.velocity_model_layers is None:
            self.velocity_model_layers = [layer]
        else:
            self.velocity_model_layers.append(layer)

    def gen_1d_model(self, z_min, z_max, spacing):
        # sort the layers to ensure the layers are properly ordered
        z = []
        v = []
        for layer in self.velocity_model_layers:
            z.append(layer.z_top)
            v.append(layer.value_top)

        if np.max(z) < z_max:
            i_z_max = np.argmax(z)
            v_z_max = v[i_z_max]

            z.append(z_max)
            v.append(v_z_max)

        if np.min(z) > z_min:
            i_z_min = np.argmin(z)
            v_z_min = v[i_z_min]

            z.append(z_min)
            v.append(v_z_min)

        i_sort = np.argsort(z)

        z = np.array(z)
        v = np.array(v)

        z = z[i_sort]
        v = v[i_sort]

        z_interp = np.arange(z_min, z_max, spacing[2])
        v_interp = np.interp(z_interp, z, v)

        return z_interp, v_interp

    def gen_3d_grid(self, network_code, dims, origin, spacing):
        model_grid_3d = VelocityGrid3D.from_layered_model(self,
                                                          network_code,
                                                          dims, origin,
                                                          spacing)
        return model_grid_3d

    def plot(self, z_min, z_max, spacing, *args, **kwargs):
        """
        Plot the 1D velocity model
        :param z_min: lower limit of the model
        :param z_max: upper limit of the model
        :param spacing: plotting resolution in z
        :return: matplotlib axis
        """

        z_interp, v_interp = self.gen_1d_model(z_min, z_max, spacing)

        x_label = None
        if self.phase == 'P':
            x_label = 'P-wave velocity'
        elif self.phase == 'S':
            x_label = 's-wave velocity'

        if self.grid_units == 'METER':
            units = 'm'
        else:
            units = 'km'

        y_label = f'z [{units}]'
        ax = plt.axes()
        ax.plot(v_interp, z_interp, *args, **kwargs)
        plt.xlabel(x_label)
        plt.ylabel(y_label)

        ax.set_aspect(2)

        plt.tight_layout()

        return ax


class VelocityGrid3D(NLLocGrid):

    def __init__(self, network_code, data_or_dims, origin, spacing,
                 phase='P', value=0, float_type=__default_float_type__,
                 model_id=None, **kwargs):

        self.network_code = network_code

        if (type(spacing) is int) | (type(spacing) is float):
            spacing = [spacing, spacing, spacing]

        super().__init__(data_or_dims, origin, spacing, phase,
                         value=value, grid_type='VELOCITY_METERS',
                         grid_units='METER', float_type=float_type,
                         model_id=model_id)

    @staticmethod
    def get_base_name(network_code, phase):
        """
        return the base name given a network code and a phase
        :param network_code: Code of the network
        :type network_code: str
        :param phase: Phase, either P or S
        :type phase: str either 'P' or 'S'
        :return: the base name
        """
        validate_phase(phase)
        return f'{network_code.upper()}.{phase.upper()}.mod'

    @classmethod
    def from_layered_model(cls, layered_model, network_code, dims, origin,
                           spacing, **kwargs):
        """
        Generating a 3D grid model from
        :param network_code:
        :param layered_model:
        :param dims:
        :param origin:
        :param spacing:
        :param kwargs:
        :return:
        """

        z_min = origin[-1]
        z_max = z_min + spacing[-1] * dims[-1]

        z_interp, v_interp = layered_model.gen_1d_model(z_min, z_max,
                                                        spacing)

        data = np.zeros(dims)

        for i, v in enumerate(v_interp):
            data[:, :, i] = v_interp[i]

        return cls(network_code, data, origin, spacing,
                   phase=layered_model.phase,
                   float_type=layered_model.float_type,
                   model_id=layered_model.model_id, **kwargs)

    def to_slow_lens(self):
        data = self.spacing[0] / self.data

        return NLLocGrid(data, self.origin, self.spacing,
                         self.phase, grid_type='SLOW_LEN',
                         grid_units=self.grid_units,
                         float_type=self.float_type,
                         model_id=self.model_id)

    def to_time(self, seed, seed_label, sub_grid_resolution=1,
                *args, **kwargs):
        """
        Eikonal solver based on scikit fast marching solver
        :param seed: numpy array location of the seed or origin of useis wave
         in model coordinates
        (usually location of a station or an event)
        :type seed: numpy array
        :param seed_label: seed label (name of station)
        :type seed_label: basestring
        :param sub_grid_resolution: resolution of the grid around the seed.
        Propagating the wavefront on a denser grid around the seed,
        significantly improves the travel time accuracy.
        :rtype: TTGrid
        """

        if not self.in_grid(seed):
            logger.warning(f'{seed_label} is outside the grid. '
                           f'The travel time grid will not be calculated')
            return

        origin = self.origin
        shape = self.shape
        spacing = self.spacing

        extent = (4 * spacing / sub_grid_resolution) * 1.2 \
                 + sub_grid_resolution

        x_i = np.arange(0, extent[0])
        y_i = np.arange(0, extent[1])
        z_i = np.arange(0, extent[2])

        x_i = x_i - np.mean(x_i) + seed[0]
        y_i = y_i - np.mean(y_i) + seed[1]
        z_i = z_i - np.mean(z_i) + seed[2]

        X_i, Y_i, Z_i = np.meshgrid(x_i, y_i, z_i, indexing='ij')

        coords = np.array([X_i.ravel(), Y_i.ravel(), Z_i.ravel()]).T

        vel = self.interpolate(coords, grid_coordinate=False).reshape(
            X_i.shape)

        phi = np.ones_like(X_i)
        phi[int(np.floor(len(x_i) / 2)), int(np.floor(len(y_i) / 2)),
            int(np.floor(len(z_i) / 2))] = 0

        tt_tmp = skfmm.travel_time(phi, vel, dx=sub_grid_resolution)

        tt_tmp_grid = TTGrid(self.network_code, tt_tmp, [x_i[0], y_i[0],
                                                         z_i[0]],
                             [sub_grid_resolution] * 3, seed, seed_label,
                             phase=self.phase, float_type=self.float_type,
                             model_id=self.model_id,
                             grid_units=self.grid_units)

        data = self.data

        xe = origin[0] + np.arange(0, shape[0], 1) * spacing[0]
        ye = origin[1] + np.arange(0, shape[1], 1) * spacing[1]
        ze = origin[2] + np.arange(0, shape[2], 1) * spacing[2]

        Xe, Ye, Ze = np.meshgrid(xe, ye, ze, indexing='ij')

        coords = np.array([Xe.ravel(), Ye.ravel(), Ze.ravel()])

        corner1 = np.array([np.min(x_i), np.min(y_i), np.min(z_i)])
        corner2 = np.array([np.max(x_i), np.max(y_i), np.max(z_i)])

        test = ((coords[0, :] >= corner1[0]) & (coords[0, :] <= corner2[0]) &
                (coords[1, :] >= corner1[1]) & (coords[1, :] <= corner2[1]) &
                (coords[2, :] >= corner1[2]) & (coords[2, :] <= corner2[2]))

        Xe_grid = Xe.ravel()[test]
        Ye_grid = Ye.ravel()[test]
        Ze_grid = Ze.ravel()[test]

        X = np.array([Xe_grid, Ye_grid, Ze_grid]).T

        tt_interp = tt_tmp_grid.interpolate(X, grid_coordinate=False, order=3)

        bias = np.max(tt_interp)

        phi_out = np.ones_like(Xe).ravel()
        phi_out[test] = tt_interp - bias

        phi_out = phi_out.reshape(Xe.shape)

        tt_out = skfmm.travel_time(phi_out, data, dx=spacing)

        # tt_out = tt_out.ravel() + bias
        tt_out = tt_out.ravel() + bias
        tt_out[test] = tt_interp
        tt_out = tt_out.reshape(Xe.shape)

        tt_out_grid = TTGrid(self.network_code, tt_out, self.origin,
                             self.spacing, seed, seed_label, phase=self.phase,
                             float_type=self.float_type,
                             model_id=self.model_id,
                             grid_units=self.grid_units)

        tt_out_grid.data -= tt_out_grid.interpolate(seed.T,
                                                    grid_coordinate=False,
                                                    order=3)

        return tt_out_grid

    def to_time_multi_threaded(self, seeds, seed_labels, cpu_utilisation=0.9,
                               *args, **kwargs):
        """
        Multithreaded version of the Eikonal solver
        based on scikit fast marching solver
        :param seeds: array of seed
        :type seeds: np.array
        :param seed_labels: array of seed_labels
        :type seed_labels: np.array
        :param cpu_utilisation: fraction of the cpu core to be used for the
        processing task (between 0 and 1)
        :type cpu_utilisation: float between 0 and 1
        :param args: arguments to be passed directly to skfmm.travel_time
        function
        :param kwargs: keyword arguments to be passed directly to
        skfmm.travel_time function
        :return: a travel time grid ensemble
        :rtype: TravelTimeEnsemble
        """

        num_threads = int(np.ceil(cpu_utilisation * __cpu_count__))
        # ensuring that the number of threads is comprised between 1 and
        # __cpu_count__
        num_threads = np.max([np.min([num_threads, __cpu_count__]), 1])

        data = []
        for seed, seed_label in zip(seeds, seed_labels):
            if not self.in_grid(seed):
                logger.warning(f'{seed_label} is outside the grid. '
                               f'The travel time grid will not be calculated')
                continue
            data.append((seed, seed_label))

        with Pool(num_threads) as pool:
            results = pool.starmap(self.to_time, data)

        tt_grid_ensemble = TravelTimeEnsemble(results)

        return tt_grid_ensemble

    def write(self, path='.'):

        base_name = self.base_name
        super().write(base_name, path=path)

    @property
    def base_name(self):
        return self.get_base_name(self.network_code, self.phase)


class VelocityGridEnsemble:
    def __init__(self, p_velocity_grid, s_velocity_grid):
        """

        :param p_velocity_grid: p-wave 3D velocity grid
        :type p_velocity_grid: VelocityGrid3D
        :param s_velocity_grid: s-wave 3D velocity grid
        :type s_velocity_grid: VelocityGrid3D
        """

        self.p_velocity_grid = p_velocity_grid
        self.s_velocity_grid = s_velocity_grid
        self.__i__ = 0

    def __getitem__(self, item):
        if item.upper() == 'P':
            return self.p_velocity_grid

        elif item.upper() == 'S':
            return self.s_velocity_grid

        else:
            raise ValueError(f'{item} is not a valid key. '
                             f'The key value must either be "P" or "S"')

    def __iter__(self):
        self.__i__ = 0
        return self

    def __next__(self):
        if self.__i__ < 2:
            if self.__i__ == '0':
                return self.p_velocity_grid
            elif self.__i__ == '1':
                return self.s_velocity_grid
        else:
            raise StopIteration

    # @property
    # def keys(self):
    #     return ['P', 'S']

    def keys(self):
        return ['P', 'S']

    def write(self, path='.'):
        for key in self.keys():
            self[key].write(path=path)

    def to_time_multi_threaded(self, seeds, seed_labels, cpu_utilisation=0.9,
                               *args, **kwargs):

        tt_grid_ensemble = TravelTimeEnsemble([])

        for key in self.keys():
            tt_grids = self[key].to_time_multi_threaded(seeds, seed_labels,
                                                        cpu_utilisation=
                                                        cpu_utilisation,
                                                        *args, **kwargs)

            tt_grid_ensemble += tt_grids

        return tt_grid_ensemble


class SeededGrid(NLLocGrid):
    """
    container for seeded grids (e.g., travel time, azimuth and take off angle)
    """
    __valid_grid_type__ = ['TIME', 'TIME2D', 'ANGLE', 'ANGLE2D']

    def __init__(self, network_code, data_or_dims, origin, spacing, seed,
                 seed_label, phase='P', value=0,
                 grid_units=__default_grid_units__,
                 grid_type='TIME', float_type="FLOAT", model_id=None):
        self.seed = seed
        self.seed_label = seed_label
        self.network_code = network_code

        if grid_type not in self.__valid_grid_type__:
            raise ValueError()
        self.grid_type = grid_type

        super().__init__(data_or_dims, origin, spacing,
                         phase=phase, value=value,
                         grid_type='TIME', grid_units=grid_units,
                         float_type=float_type, model_id=model_id)

    def __repr__(self):
        line = f'Travel Time Grid\n' \
               f'    origin        : {self.origin}\n' \
               f'    spacing       : {self.spacing}\n'  \
               f'    dimensions    : {self.shape}\n' \
               f'    seed label    : {self.seed_label}\n' \
               f'    seed location : {self.seed}'
        return line

    @classmethod
    def get_base_name(cls, network_code, phase, seed_label, grid_type):
        validate_phase(phase)
        if grid_type not in cls.__valid_grid_type__:
            raise ValueError(f'{grid_type} is not a valid grid type')

        base_name = f'{network_code}.{phase}.{seed_label}.' \
                    f'{grid_type.lower()}'
        return base_name

    @property
    def base_name(self):
        base_name = self.get_base_name(self.network_code, self.phase,
                                       self.seed_label, self.grid_type)
        return base_name

    def write(self, path='.'):
        base_name = self.base_name
        self._write_grid_data(base_name, path=path)
        self._write_grid_header(base_name, path=path, seed=self.seed,
                                seed_label=self.seed_label,
                                seed_units=self.grid_units)
        self._write_grid_model_id(base_name, path=path)


class TTGrid(SeededGrid):
    def __init__(self, network_code, data_or_dims, origin, spacing, seed,
                 seed_label, phase='P', value=0, float_type="FLOAT",
                 model_id=None, grid_units=__default_grid_units__):

        super().__init__(network_code, data_or_dims, origin, spacing, seed,
                         seed_label, phase=phase, value=value,
                         grid_type='TIME', float_type=float_type,
                         model_id=model_id, grid_units=grid_units)

    def to_azimuth(self):
        """
        This function calculate the take off angle and azimuth for every
        grid point given a travel time grid calculated using an Eikonal solver
        :return: azimuth and takeoff angles grids
        .. Note: The convention for the takeoff angle is that 0 degree is down.
        """

        gds_tmp = np.gradient(self.data)
        gds = [-gd for gd in gds_tmp]

        azimuth = np.arctan2(gds[0], gds[1]) * 180 / np.pi
        # azimuth is zero northwards

        return AngleGrid(self.network_code, azimuth, self.origin, self.spacing,
                         self.seed, self.seed_label, 'AZIMUTH',
                         phase=self.phase, float_type=self.float_type,
                         model_id=self.model_id, grid_units=self.grid_units)

    def to_takeoff(self):
        gds_tmp = np.gradient(self.data)
        gds = [-gd for gd in gds_tmp]

        hor = np.sqrt(gds[0] ** 2 + gds[1] ** 2)
        takeoff = np.arctan2(hor, -gds[2]) * 180 / np.pi
        # takeoff is zero pointing down
        return AngleGrid(self.network_code, takeoff, self.origin, self.spacing,
                         self.seed, self.seed_label, 'TAKEOFF',
                         phase=self.phase, float_type=self.float_type,
                         model_id=self.model_id, grid_units=self.grid_units)

    def to_azimuth_point(self, coord, grid_coordinate=False, mode='nearest',
                         order=1, **kwargs):
        """
        calculate the azimuth angle at a particular point on the grid for a
        given seed location
        :param coord: coordinates at which to calculate the takeoff angle
        :param grid_coordinate: true if the coordinates are expressed in
        grid space (indices can be fractional) as opposed to model space
        (x, y, z)
        :param mode: interpolation mode
        :param order: interpolation order
        :return: takeoff angle at the location coord
        """

        return self.to_azimuth().interpolate(coord,
                                             grid_coordinate=grid_coordinate,
                                             mode=mode, order=order, **kwargs)

    def to_takeoff_point(self, coord, grid_coordinate=False, mode='nearest',
                         order=1, **kwargs):
        """
        calculate the takeoff angle at a particular point on the grid for a
        given seed location
        :param coord: coordinates at which to calculate the takeoff angle
        :param grid_coordinate: true if the coordinates are expressed in
        grid space (indices can be fractional) as opposed to model space
        (x, y, z)
        :param mode: interpolation mode
        :param order: interpolation order
        :return: takeoff angle at the location coord
        """
        return self.to_takeoff().interpolate(coord,
                                             grid_coordinate=grid_coordinate,
                                             mode=mode, order=order, **kwargs)

    def ray_tracer(self, start, grid_coordinate=False, max_iter=1000,
                   arrival_id=None):
        """
        This function calculates the ray between a starting point (start) and an
        end point, which should be the seed of the travel_time grid, using the
        gradient descent method.
        :param start: the starting point (usually event location)
        :type start: tuple, list or numpy.array
        :param grid_coordinate: true if the coordinates are expressed in
        grid space (indices can be fractional) as opposed to model space
        (x, y, z)
        :param max_iter: maximum number of iteration
        :param arrival_id: id of the arrival associated to the ray if
        applicable
        :rtype: numpy.array
        """

        return ray_tracer(self, start, grid_coordinate=grid_coordinate,
                          max_iter=max_iter, arrival_id=arrival_id,
                          earth_model_id=self.model_id)

    @classmethod
    def from_velocity(cls, seed, seed_label, velocity_grid):
        return velocity_grid.eikonal(seed, seed_label)

    def write(self, path='.'):
        return super().write(path=path)


class TravelTimeEnsemble:
    def __init__(self, travel_time_grids):
        """
        Combine a list of travel time grids together providing meta
        functionality (multi-threaded ray tracing, sorting, travel-time
        calculation for a specific location etc.). It is assumed that
        all grids are compatible, i.e., that all the grids have the same
        origin, spacing and dimensions.
        :param travel_time_grids: a list of TTGrid objects
        """

        self.travel_time_grids = travel_time_grids
        self.__i__ = 0

        for tt_grid in self.travel_time_grids:
            try:
                assert tt_grid.check_compatibility(travel_time_grids[0])
            except:
                raise AssertionError('grids are not all compatible')

    def __len__(self):
        return len(self.travel_time_grids)

    def __add__(self, other):
        for travel_time_grid in other.travel_time_grids:
            self.travel_time_grids.append(travel_time_grid)

        return TravelTimeEnsemble(self.travel_time_grids)

    def __iter__(self):
        self.__i__ = 0
        return self

    def __next__(self):
        if self.__i__ < len(self):
            result = self.travel_time_grids[self.__i__]
            self.__i__ += 1
            return result
        else:
            raise StopIteration

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.travel_time_grids[item]
        if isinstance(item, str):
            tt_grid_out = None
            for travel_time_grid in self.travel_time_grids:
                if travel_time_grid.seed_label == item:
                    return travel_time_grid

            raise KeyError(f'{item} not found')

    def __repr__(self):
        line = f'Number of travel time grids: {len(self)}'
        return line

    @classmethod
    def from_files(cls, path):
        """
        create a travel time ensemble from files located in a directory
        :param path: the base path to the directory containing the travel time
        files.
        :return:
        """
        tt_grids = []
        for fle in Path(path).glob('*time*.hdr'):
            path = fle.parent
            base_name = '.'.join(fle.name.split('.')[:-1])
            tt_grid = read_grid(str(base_name), path=str(path),
                                float_type=__default_float_type__)
            tt_grids.append(tt_grid)

        return cls(tt_grids)

    def select(self, seed_labels: Optional[list] = None,
               phase: Optional[list] = None):
        """
        return the a list of grid corresponding to seed_labels.
        :param seed_labels: seed labels of the travel time grids to return
        :param phase: the phase {'P' or 'S'}, both if None.
        :return: a list of travel time grids
        :rtype: TravelTimeEnsemble
        """

        if (seed_labels is None) and (phase is None):
            return self

        tmp = []
        if seed_labels is None:
            seed_labels = np.unique(self.seed_labels)

        if phase is None:
            phase = ['P', 'S']

        returned_grids = []
        for travel_time_grid in self.travel_time_grids:
            if travel_time_grid.seed_label in seed_labels:
                if travel_time_grid.phase in phase:
                    returned_grids.append(travel_time_grid)

        return TravelTimeEnsemble(returned_grids)

    def sort(self, ascending=True):
        """
        sorting the travel time grid by seed_label
        :param ascending: if true the grids are sorted in ascending order
        :param ascending: bool
        :return: sorted travel time grids.
        :rtype: TravelTimeEnsemble
        """

        i = np.sort(self.seed_labels)

        if not ascending:
            i = i[-1::-1]

        sorted_tt_grids = np.array(self.travel_time_grids)[i]

        return TravelTimeEnsemble(sorted_tt_grids)

    def travel_time(self, seed, grid_coordinate: bool=False,
                    seed_labels: Optional[list]=None,
                    phase: Optional[list]=None):
        """
        calculate the travel time at a specific point for a series of site
        ids
        :param seed: travel time seed
        :param grid_coordinate: true if the coordinates are expressed in
        grid space (indices can be fractional) as opposed to model space
        (x, y, z)
        :param seed_labels: a list of sites from which to calculate the
        travel time.
        :param phase: a list of phases for which the travel time need to be
        calculated
        :return: a list of dictionary containing the travel time and site id
        """

        if isinstance(seed, list):
            seed = np.array(seed)

        if not self.travel_time_grids[0].in_grid(seed):
            raise ValueError('seed is outside the grid')

        if grid_coordinate:
            seed = self.travel_time_grids[0].transform_from(seed)

        tt_grids = self.select(seed_labels=seed_labels, phase=phase)

        tts = []
        labels = []
        phases = []
        for tt_grid in tt_grids:
            labels.append(tt_grid.seed_label)
            tts.append(tt_grid.interpolate(seed.T, grid_coordinate=False)[0])
            phases.append(tt_grid.phase)

        tts_dict = {}
        for phase in np.unique(phases):
            tts_dict[phase] = {}

        for label, tt, phase in zip(labels, tts, phases):
            tts_dict[phase][label] = tt

        return tts_dict

    def ray_tracer(self, start, seed_labels=None, multithreading=False,
                   cpu_utilisation=0.9, grid_coordinate=False, max_iter=1000):
        """

        :param start: origin of the ray, usually the location of an event
        :param seed_labels: a list of seed labels
        :param grid_coordinate: true if the coordinates are expressed in
        grid space (indices can be fractional) as opposed to model space
        (x, y, z)
        :param multithreading: if True use multithreading
        :param max_iter: maximum number of iteration
        :param cpu_utilisation: fraction of core to use, between 0 and 1.
        The number of core to be use is bound between 1 and the total number of
        cores
        :return: a list of ray
        :rtype: list
        """

        travel_time_grids = self.select(seed_labels=seed_labels)

        if multithreading:

            kwargs = {'grid_coordinate': grid_coordinate,
                      'max_iter': max_iter}

            ray_tracer_func = partial(ray_tracer, **kwargs)

            num_threads = int(np.ceil(cpu_utilisation * __cpu_count__))
            # ensuring that the number of threads is comprised between 1 and
            # __cpu_count__
            num_threads = np.max([np.min([num_threads, __cpu_count__]), 1])

            data = []
            for travel_time_grid in travel_time_grids:
                data.append((travel_time_grid, start))

            with Pool(num_threads) as pool:
                results = pool.starmap(ray_tracer_func, data)

        else:
            results = []
            for travel_time_grid in travel_time_grids:
                results.append(ray_tracer(travel_time_grid, start,
                                          grid_coordinate=grid_coordinate,
                                          max_iter=max_iter))

        return results

    @property
    def seeds(self):
        seeds = []
        for grid in self.travel_time_grids:
            seeds.append(grid.seed)

        return np.array(seeds)

    @property
    def seed_labels(self):
        seed_labels = []
        for grid in self.travel_time_grids:
            seed_labels.append(grid.seed_label)

        return np.array(seed_labels)

    @property
    def shape(self):
        return self.travel_time_grids[0].shape

    @property
    def origin(self):
        return self.travel_time_grids[0].origin

    @property
    def spacing(self):
        return self.travel_time_grids[0].spacing

    def write(self, path='.'):
        for travel_time_grid in self.travel_time_grids:
            travel_time_grid.write(path=path)

    def write_hdf5(self, file_name):
        hf = h5py.File(file_name, 'w')

        shape = self.shape
        origin = self.origin
        spacing = self.spacing
        ngrid = np.product(shape)
        gridlocs = gdef_to_points(shape, origin, spacing)
        hf.create_dataset('grid_locs', data=gridlocs.astype(np.float32))

        hf.attrs['shape'] = shape
        hf.attrs['origin'] = origin
        hf.attrs['spacing'] = spacing
        for phase in ['P', 'S']:
            sorted_tt_grids = self.travel_time_grids.select(phase=phase).sort()
            seeds = sorted_tt_grids.seeds
            seed_labels = np.array(sorted_tt_grids.seed_labels)
            nsta = len(sorted_tt_grids)
            tts = np.zeros((nsta, ngrid), dtype=np.float32)

            for i, tt_grid in enumerate(sorted_tt_grids):
                tts[i] = tt_grid.data.reshape(ngrid)

            seeds = np.array(seeds).astype(np.float32)

            hf.create_dataset(f'tt{phase.lower()}',
                              data=tts)

        hf.create_dataset('locations', data=seeds)
        gdef = np.concatenate((shape, origin, [spacing])).astype(np.int32)
        hf.create_dataset('grid_def', data=gdef)
        hf.create_dataset('stations', data=seed_labels.astype('S4'))
        hf.close()


class AngleGrid(SeededGrid):
    def __init__(self, network_code, data_or_dims, origin, spacing, seed,
                 seed_label, angle_type, phase='P', value=0,
                 float_type="FLOAT", model_id=None,
                 grid_units=__default_grid_units__):

        self.angle_type = angle_type
        super().__init__(network_code, data_or_dims, origin, spacing, seed,
                         seed_label, phase=phase, value=value,
                         grid_type='ANGLE', float_type=float_type,
                         model_id=model_id)

    def write(self, path='.'):
        pass


def gdef_to_points(shape, origin, spacing):
    maxes = origin + shape * spacing
    x = np.arange(origin[0], maxes[0], spacing).astype(np.float32)
    y = np.arange(origin[1], maxes[1], spacing).astype(np.float32)
    z = np.arange(origin[2], maxes[2], spacing).astype(np.float32)
    points = np.zeros((np.product(shape), 3), dtype=np.float32)
    ix = 0

    for xv in x:
        for yv in y:
            for zv in z:
                points[ix] = [xv, yv, zv]
                ix += 1

    return points




# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
# Filename: grid.py
#  Purpose: plugin for reading and writing GridData object into various format 
#   Author: uquake development team
#    Email: devs@uquake.org
#
# Copyright (C) 2016 uquake development team
# --------------------------------------------------------------------
"""
plugin for reading and writing GridData object into various format 

:copyright:
    uquake development team (devs@uquake.org)
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

import numpy as np


def read_pickle(filename, protocol=-1, **kwargs):
    """
    read grid saved in PICKLE format into a GridData object
    :param filename: full path to the filename
    :type filename: str
    :rtype: ~uquake.core.data.grid.Grid
    """
    import pickle
    return pickle.load(open(filename, 'rb'))


def write_pickle(grid, filename, protocol=-1, **kwargs):
    """
    write a GridData object to disk in pickle (.pickle or .npy extension)
    format
    using the pickle module
    :param grid: grid to be saved
    :type grid: ~uquake.core.data.grid.GridData
    :param filename: full path to file with extension
    :type filename: str
    :param protocol: pickling protocol level
    :type protocol: int
    """
    import pickle
    with open(filename, 'wb') as of:
        pickle.dump(grid, of, protocol=protocol)

    return True


def read_hdf5(filename, **kwargs):
    """
    read a grid file in hdf5 into a uquake.core.data.grid.GridCollection
    object
    :param filename: filename
    :param kwargs: additional keyword argument passed from wrapper.
    :return: uquake.core.data.grid.GridCollection
    """
    pass


def write_csv(grid, filename, **kwargs):
    """
    Write a GridData object to disk in Microquake csv format
    :param grid: grid to be saved
    :param filename: full path to file with extension
    :return:
    """
    data = grid.data
    shape = grid.shape
    origin = grid.origin
    spacing = grid.spacing

    v = grid.get_grid_point_coordinates()

    flat_data = data.reshape(np.product(shape))

    with open(filename, 'w') as f_out:
        if data.ndim == 3:
            f_out.write('uquake grid\n')
            f_out.write(f'spacing: {spacing}')
            f_out.write(f'origin: {origin}')
            f_out.write(f'shape: {shape}')
            f_out.write('x,y,z,value\n')

            for k in range(grid.shape[2]):
                for j in range(grid.shape[1]):
                    for i in range(grid.shape[0]):
                        f_out.write(f'{v[0][i]},{v[1][j]},{v[2][k]},'
                                    f'{grid.data[i, j, k]}\n')

        elif data.ndim == 2:
            f_out.write('uquake grid\n')
            f_out.write(f'spacing: {spacing}')
            f_out.write(f'origin: {origin}')
            f_out.write(f'shape: {shape}')
            f_out.write('x,y,value\n')
            for j in range(grid.shape[1]):
                for i in range(grid.shape[0]):
                    f_out.write(f'{v[0][i]},{v[0][j]},'
                                f'{grid.data[i, j]}\n')

    return True


def read_csv(filename, *args, **kwargs):
    """
    Read a grid save in uquake CSV format
    :param filename: path to file
    :param args:
    :param kwargs:
    :return:
    """
    pass


def write_vtk(grid, filename, **kwargs):
    """
    write a GridData object to disk in VTK format (Paraview, MayaVi2,
    etc.) using
    the pyevtk module.
    param filename: full path to file with the extension. Note that the
    extension for vtk image data (grid data) is usually .vti. 
    :type filename; str
    :param grid: grid to be saved
    :type grid: ~uquake.core.data.grid.GridData
    .. NOTE:
        see the imageToVTK function from the pyevtk.hl module for more
        information on possible additional paramter.
    """
    import vtk

    if filename[-4:] in ['.vti', '.vtk']:
        filename = filename[:-4]

    image_data = vtk.vtkImageData()
    image_data.SetDimensions(grid.shape)
    image_data.SetSpacing(grid.spacing)
    image_data.SetOrigin(grid.origin)
    image_data.AllocateScalars(vtk.VTK_FLOAT, 1)

    if grid.ndim == 3:
        for z in range(grid.shape[2]):
            for y in range(grid.shape[1]):
                for x in range(grid.shape[0]):
                    image_data.SetScalarComponentFromFloat(x, y, z, 0,
                                                           grid.data[x, y, z])

    if grid.ndim == 2:
        for y in range(grid.shape[1]):
            for x in range(grid.shape[0]):
                image_data.SetScalarComponentFromFloat(x, y, 0,
                                                       grid.data[x, y])

    writer = vtk.vtkXMLImageDataWriter()
    writer.SetFileName(f'{filename}.vti')
    writer.SetInputData(image_data)
    writer.Write()
    return True


def read_vtk(filename, *args, **kwargs):
    pass